import requests

from allauth.socialaccount.providers.oauth2.views import (
    OAuth2Adapter,
    OAuth2CallbackView,
    OAuth2LoginView,
)


class TUMOAuth2Adapter(OAuth2Adapter):
    provider_id = "tum"
    access_token_url = "https://campus.tum.de/tumonline/wbOAuth2.token"
    authorize_url = "https://campus.tum.de/tumonline/wbOAuth2.authorize"
    profile_url = "https://campus.tum.de/tumonline/!co_loc_wsoa2user.userinfo"

    def complete_login(self, request, app, token, **kwargs):
        extra_data = self.get_user_info(token)
        login = self.get_provider().sociallogin_from_response(
            request, extra_data
        )
        return login

    def get_user_info(self, token):
        auth_header = f"Bearer {token.token}"
        headers = {
            "Accept": "application/json",
            "Authorization": auth_header,
            "accept": "application/json",
        }
        resp = requests.get(self.profile_url, headers=headers)
        resp.raise_for_status()
        return resp.json()


oauth2_login = OAuth2LoginView.adapter_view(TUMOAuth2Adapter)
oauth2_callback = OAuth2CallbackView.adapter_view(TUMOAuth2Adapter)



from selenium.webdriver.common.by import By
from channels.testing import ChannelsLiveServerTestCase
from testing.selenium_helper import SeleniumHelper


class AccessTest(SeleniumHelper, ChannelsLiveServerTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.wait_time = 10

    def setUp(self):
        self.base_url = self.live_server_url
        return super().setUp()

    def test_ie11(self):
        driver_data = self.get_drivers(
            1,
            user_agent=(
                "Mozilla/5.0 (compatible, MSIE 11, Windows NT 6.3; "
                "Trident/7.0; rv:11.0) like Gecko"
            ),
        )
        driver = driver_data["drivers"][0]
        driver.implicitly_wait(driver_data["wait_time"])
        driver.get(self.base_url + "/")
        h1 = driver.find_element(By.CSS_SELECTOR, "h1")
        self.assertEqual(h1.text, "Browser not supported")
        driver.quit()

    def test_robots(self):
        driver_data = self.get_drivers(
            1,
            user_agent=(
                "Mozilla/5.0 (compatible; Googlebot/2.1; "
                "+http://www.google.com/bot.html)"
            ),
        )
        driver = driver_data["drivers"][0]
        driver.implicitly_wait(driver_data["wait_time"])
        driver.get(self.base_url + "/robots.txt")
        body = driver.find_element(By.CSS_SELECTOR, "body")
        self.assertEqual(
            body.text,
            "User-agent: *\nDisallow: /document/\nDisallow: /bibliography/",
        )
        driver.quit()

    def test_hello(self):
        driver_data = self.get_drivers(1)
        driver = driver_data["drivers"][0]
        driver.implicitly_wait(driver_data["wait_time"])
        driver.get(self.base_url + "/hello-fiduswriter")
        body = driver.find_element(By.CSS_SELECTOR, "body")
        self.assertEqual(body.text, "Hello from Fidus Writer")
        driver.quit()



import time
import os
import bleach
import json
from copy import deepcopy

from django.core import serializers
from django.http import HttpResponse, JsonResponse, HttpRequest
from django.contrib.auth.decorators import login_required
from django.views.decorators.http import require_POST
from django.db import transaction
from django.core.files import File
from django.utils.translation import gettext as _
from django.conf import settings
from django.db.models import F, Q
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth import get_user_model

from document.models import (
    Document,
    AccessRight,
    DocumentRevision,
    DocumentTemplate,
    CAN_UPDATE_DOCUMENT,
    CAN_COMMUNICATE,
    FW_DOCUMENT_VERSION,
)
from usermedia.models import DocumentImage, Image
from bibliography.models import Entry
from document.helpers.serializers import PythonWithURLSerializer
from bibliography.views import serializer
from style.models import DocumentStyle, DocumentStyleFile, ExportTemplate
from base.decorators import ajax_required
from base.helpers.ws import get_url_base
from user.models import UserInvite
from . import emails
from user.helpers import Avatars


@login_required
@ajax_required
@require_POST
def get_documentlist_extra(request):
    response = {}
    status = 200
    ids = request.POST["ids"].split(",")
    docs = (
        Document.objects.filter(
            Q(owner=request.user) | Q(accessright__user=request.user)
        )
        .filter(id__in=ids)
        .prefetch_related("documentimage_set")
    )
    response["documents"] = []
    for doc in docs:
        images = {}
        for image in doc.documentimage_set.all():
            images[image.image.id] = {
                "added": image.image.added,
                "checksum": image.image.checksum,
                "file_type": image.image.file_type,
                "height": image.image.height,
                "id": image.image.id,
                "image": image.image.image.url,
                "title": image.title,
                "copyright": image.copyright,
                "width": image.image.width,
            }
            if image.image.thumbnail:
                images[image.image.id]["thumbnail"] = image.image.thumbnail.url
        if "type" not in doc.content:
            doc.content = deepcopy(doc.template.content)
            if "type" not in doc.content:
                doc.content["type"] = "doc"
            if "content" not in doc.content:
                doc.content["content"] = [{type: "title"}]
            doc.save()
        response["documents"].append(
            {
                "images": images,
                "content": doc.content,
                "comments": doc.comments,
                "bibliography": doc.bibliography,
                "id": doc.id,
            }
        )
    return JsonResponse(response, status=status)


def documents_list(request):
    avatars = Avatars()
    documents = (
        Document.objects.filter(
            Q(owner=request.user) | Q(accessright__user=request.user),
            listed=True,
        )
        .defer("content", "comments", "bibliography", "doc_version", "diffs")
        .select_related("owner")
        .prefetch_related(
            "accessright_set",
            "documentrevision_set",
        )
        .distinct()
        .order_by("-updated")
    )
    output_list = []
    for document in documents:
        if document.owner == request.user:
            access_right = "write"
            path = document.path
        else:
            access_object = document.accessright_set.filter(
                user=request.user
            ).first()
            access_right = access_object.rights
            path = access_object.path
        if (
            request.user.is_staff
            or document.owner == request.user
            or document.accessright_set.filter(
                user=request.user, rights__in=CAN_COMMUNICATE
            ).first()
        ):
            revisions = document.documentrevision_set.all()
            revision_list = []
            for revision in revisions:
                revision_list.append(
                    {
                        "date": time.mktime(revision.date.utctimetuple()),
                        "note": revision.note,
                        "file_name": revision.file_name,
                        "pk": revision.pk,
                    }
                )
        else:
            revision_list = []
        added = time.mktime(document.added.utctimetuple())
        updated = time.mktime(document.updated.utctimetuple())
        is_owner = False
        if document.owner == request.user:
            is_owner = True
        output_list.append(
            {
                "id": document.id,
                "title": document.title,
                "path": path,
                "is_owner": is_owner,
                "owner": {
                    "id": document.owner.id,
                    "name": document.owner.readable_name,
                    "avatar": avatars.get_url(document.owner),
                },
                "added": added,
                "updated": updated,
                "rights": access_right,
                "revisions": revision_list,
            }
        )
    return output_list


@login_required
@ajax_required
@require_POST
def get_access_rights(request):
    response = {}
    status = 200
    avatars = Avatars()
    ar_qs = AccessRight.objects.filter(document__owner=request.user)
    doc_ids = request.POST.getlist("document_ids[]")
    if len(doc_ids) > 0:
        ar_qs = ar_qs.filter(document_id__in=doc_ids)
    access_rights = []
    for ar in ar_qs:
        if ar.holder_type.model == "user":
            avatar = avatars.get_url(ar.holder_obj)
        else:
            avatar = None
        access_rights.append(
            {
                "document_id": ar.document.id,
                "rights": ar.rights,
                "holder": {
                    "id": ar.holder_id,
                    "type": ar.holder_type.model,
                    "name": ar.holder_obj.readable_name,
                    "avatar": avatar,
                },
            }
        )
    response["access_rights"] = access_rights
    return JsonResponse(response, status=status)


@login_required
@ajax_required
@require_POST
@transaction.atomic
def save_access_rights(request):
    User = get_user_model()
    response = {}
    doc_ids = json.loads(request.POST["document_ids"])
    rights = json.loads(request.POST["access_rights"])
    for doc_id in doc_ids:
        doc = Document.objects.filter(pk=doc_id, owner=request.user).first()
        if not doc:
            continue
        for right in rights:
            holder_selector = "%(type)s__id" % {
                "type": right["holder"]["type"]
            }
            if right["rights"] == "delete":
                # Status 'delete' means the access right is marked for
                # deletion.
                AccessRight.objects.filter(
                    **{
                        "document_id": doc_id,
                        holder_selector: right["holder"]["id"],
                    }
                ).delete()
            else:
                owner = request.user.readable_name
                link = HttpRequest.build_absolute_uri(
                    request, doc.get_absolute_url()
                )
                access_right = AccessRight.objects.filter(
                    **{
                        "document_id": doc_id,
                        holder_selector: right["holder"]["id"],
                    }
                ).first()
                document_title = doc.title
                if access_right:
                    if access_right.rights != right["rights"]:
                        access_right.rights = right["rights"]
                        if right["holder"]["type"] == "user":
                            collaborator = User.objects.get(
                                id=right["holder"]["id"]
                            )
                            collaborator_name = collaborator.readable_name
                            collaborator_email = collaborator.email
                            emails.send_share_notification(
                                document_title,
                                owner,
                                link,
                                collaborator_name,
                                collaborator_email,
                                right["rights"],
                                True,
                            )
                else:
                    # Make the shared path "/filename" or ""
                    path = "/%(last_path_part)s" % {
                        "last_path_part": doc.path.split("/").pop()
                    }
                    if len(path) == 1:
                        path = ""
                    if right["holder"]["type"] == "userinvite":
                        holder = UserInvite.objects.filter(
                            id=right["holder"]["id"]
                        ).first()
                    else:
                        holder = User.objects.filter(
                            id=right["holder"]["id"]
                        ).first()
                    if not holder:
                        continue
                    access_right = AccessRight.objects.create(
                        document_id=doc_id,
                        holder_obj=holder,
                        rights=right["rights"],
                        path=path,
                    )
                    if right["holder"]["id"] == "user":
                        collaborator_name = holder.readable_name
                        collaborator_email = holder.email
                        emails.send_share_notification(
                            document_title,
                            owner,
                            link,
                            collaborator_name,
                            collaborator_email,
                            right["rights"],
                            False,
                        )
                access_right.save()
    status = 201
    return JsonResponse(response, status=status)


@login_required
@ajax_required
@require_POST
def get_documentlist(request):
    response = {}
    status = 200
    response["documents"] = documents_list(request)
    response["contacts"] = []
    avatars = Avatars()
    for contact in request.user.contacts.all():
        contact_object = {
            "id": contact.id,
            "name": contact.readable_name,
            "username": contact.get_username(),
            "avatar": avatars.get_url(contact),
            "type": "user",
        }
        response["contacts"].append(contact_object)
    for contact in request.user.invites_by.all():
        contact_object = {
            "id": contact.id,
            "name": contact.username,
            "username": contact.username,
            "avatar": None,
            "type": "userinvite",
        }
        response["contacts"].append(contact_object)
    serializer = PythonWithURLSerializer()
    doc_styles = serializer.serialize(
        DocumentStyle.objects.filter(
            Q(document_template__user=None)
            | Q(document_template__user=request.user)
        ),
        use_natural_foreign_keys=True,
        fields=["title", "slug", "contents", "documentstylefile_set"],
    )
    response["document_styles"] = [obj["fields"] for obj in doc_styles]
    doc_templates = DocumentTemplate.objects.filter(
        Q(user=request.user) | Q(user=None)
    ).order_by(F("user").desc(nulls_first=True))
    response["document_templates"] = {}
    for obj in doc_templates:
        response["document_templates"][obj.import_id] = {
            "title": obj.title,
            "id": obj.id,
        }
    return JsonResponse(response, status=status)


@login_required
@ajax_required
@require_POST
def delete(request):
    response = {}
    status = 200
    doc_id = int(request.POST["id"])
    document = Document.objects.get(pk=doc_id, owner=request.user)
    if document.is_deletable():
        image_ids = list(
            DocumentImage.objects.filter(document_id=doc_id).values_list(
                "image_id", flat=True
            )
        )
        document.delete()
        for image in Image.objects.filter(id__in=image_ids):
            if image.is_deletable():
                image.delete()
        response["done"] = True
    else:
        response["done"] = False
    return JsonResponse(response, status=status)


@login_required
@ajax_required
@require_POST
def move(request):
    response = {}
    status = 200
    doc_id = int(request.POST["id"])
    path = request.POST["path"]
    document = Document.objects.filter(pk=doc_id).first()
    if not document:
        response["done"] = False
    elif document.owner == request.user:
        document.path = path
        document.save(
            update_fields=[
                "path",
            ]
        )
        response["done"] = True
    else:
        access_right = AccessRight.objects.filter(
            document=document, user=request.user
        ).first()
        if not access_right:
            response["done"] = False
        else:
            access_right.path = path
            access_right.save()
            response["done"] = True
    return JsonResponse(response, status=status)


@login_required
@ajax_required
@require_POST
def create_doc(request):
    response = {}
    template_id = request.POST["template_id"]
    path = request.POST["path"]
    document_template = DocumentTemplate.objects.filter(
        Q(user=request.user) | Q(user=None), id=template_id
    ).first()
    if not document_template:
        return JsonResponse(response, status=405)
    document = Document.objects.create(
        owner_id=request.user.pk, template_id=template_id, path=path
    )
    response["id"] = document.id
    return JsonResponse(response, status=201)


@login_required
@ajax_required
@require_POST
def import_create(request):
    # First step of import: Create a document and return the id of it
    response = {}
    status = 201
    import_id = request.POST["import_id"]
    document_template = (
        DocumentTemplate.objects.filter(
            Q(user=request.user) | Q(user=None), import_id=import_id
        )
        .order_by(F("user").desc(nulls_last=True))
        .first()
    )
    if not document_template:
        # The user doesn't have this template.
        # We check whether the template exists with one of the documents
        # shared with the user. If so, we'll copy it so that we can avoid
        # having to create an entirely new template without styles or
        # exporter templates
        access_right = AccessRight.objects.filter(
            user=request.user, document__template__import_id=import_id
        ).first()
        if access_right:
            document_template = access_right.document.template
            document_styles = list(document_template.documentstyle_set.all())
            export_templates = list(document_template.exporttemplate_set.all())
            document_template.pk = None
            document_template.user = request.user
            document_template.save()
            for ds in document_styles:
                style_files = list(ds.documentstylefile_set.all())
                ds.pk = None
                ds.document_template = document_template
                ds.save()
                for sf in style_files:
                    sf.pk = None
                    sf.style = ds
                    sf.save()
            for et in export_templates:
                et.pk = None
                et.document_template = document_template
                et.save()
    if not document_template:
        template_title = request.POST["template_title"]
        counter = 0
        base_template_title = template_title
        while DocumentTemplate.objects.filter(
            Q(title=template_title), Q(user=request.user) | Q(user=None)
        ).first():
            counter += 1
            template_title = f"{base_template_title} {counter}"
        content = json.loads(request.POST["template"])
        document_template = DocumentTemplate()
        document_template.title = template_title
        document_template.import_id = import_id
        document_template.user = request.user
        document_template.content = content
        document_template.save()
        files = request.FILES.getlist("files[]")
        document_styles = json.loads(request.POST.get("document_styles"))
        export_templates = json.loads(request.POST.get("export_templates"))
        for style in document_styles:
            doc_style = DocumentStyle.objects.create(
                title=style["title"],
                slug=style["slug"],
                contents=style["contents"],
                document_template=document_template,
            )
            for filepath in style["files"]:
                filename = filepath.split("/").pop()
                file = next((x for x in files if x.name == filename), None)
                if file:
                    DocumentStyleFile.objects.create(
                        file=file, style=doc_style
                    )
        for e_template in export_templates:
            filename = e_template["file"].split("/").pop()
            file = next((x for x in files if x.name == filename), None)
            if file:
                ExportTemplate.objects.create(
                    document_template=document_template,
                    template_file=file,
                    file_type=e_template["file_type"],
                )
    path = request.POST["path"]
    if len(path):
        counter = 0
        base_path = path
        while (
            Document.objects.filter(owner=request.user, path=path).first()
            or AccessRight.objects.filter(user=request.user, path=path).first()
        ):
            counter += 1
            path = f"{base_path} {counter}"
    document = Document.objects.create(
        owner=request.user, template=document_template, path=path
    )
    response["id"] = document.id
    response["path"] = document.path
    return JsonResponse(response, status=status)


@login_required
@ajax_required
@require_POST
def import_image(request):
    # create an image for a document
    response = {}
    document = Document.objects.filter(
        owner_id=request.user.pk, id=int(request.POST["doc_id"])
    ).first()
    if document:
        status = 201
    else:
        status = 401
        return JsonResponse(response, status=status)
    checksum = int(request.POST.get("checksum", 0))
    if checksum > 0:
        image = Image.objects.filter(checksum=checksum).first()
    else:
        image = None
    if image is None:
        image = Image(
            uploader=request.user,
            image=request.FILES["image"],
            checksum=checksum,
        )
        image.save()
    doc_image = DocumentImage.objects.create(
        image=image,
        title=request.POST["title"],
        copyright=json.loads(request.POST["copyright"]),
        document=document,
    )
    response["id"] = doc_image.image.id
    return JsonResponse(response, status=status)


@login_required
@ajax_required
@require_POST
def import_doc(request):
    response = {}
    doc_id = request.POST["id"]
    # There is a doc_id, so we overwrite an existing doc rather than
    # creating a new one.
    document = Document.objects.get(id=int(doc_id))
    if (
        document.owner != request.user
        and not AccessRight.objects.filter(
            document_id=doc_id,
            user=request.user,
            rights__in=CAN_UPDATE_DOCUMENT,
        ).first()
    ):
        response["error"] = "No access to file"
        status = 403
        return JsonResponse(response, status=status)
    document.title = request.POST["title"]
    document.content = json.loads(request.POST["content"])
    document.comments = json.loads(request.POST["comments"])
    document.bibliography = json.loads(request.POST["bibliography"])
    # document.doc_version should always be the current version, so don't
    # bother about it.
    document.save()
    response["document_id"] = document.id
    response["added"] = time.mktime(document.added.utctimetuple())
    response["updated"] = time.mktime(document.updated.utctimetuple())
    status = 200
    return JsonResponse(response, status=status)


@login_required
@ajax_required
@require_POST
def upload_revision(request):
    response = {}
    status = 405
    can_save = False
    document_id = request.POST["document_id"]
    document = Document.objects.filter(id=int(document_id)).first()
    if document:
        if document.owner == request.user:
            can_save = True
        else:
            access_rights = AccessRight.objects.filter(
                document=document, user=request.user
            )
            if len(access_rights) > 0 and access_rights[0].rights == "write":
                can_save = True
    if can_save:
        status = 201
        revision = DocumentRevision()
        revision.file_object = request.FILES["file"]
        revision.file_name = request.FILES["file"].name
        revision.note = request.POST["note"]
        revision.document_id = document_id
        revision.save()
    return JsonResponse(response, status=status)


# Download a revision that was previously uploaded
@login_required
def get_revision(request, revision_id):
    revision = DocumentRevision.objects.filter(pk=int(revision_id)).first()
    if revision and (
        request.user.is_staff
        or revision.document.owner == request.user
        or AccessRight.objects.filter(
            document=revision.document,
            user=request.user,
            rights__in=CAN_COMMUNICATE,
        ).first()
    ):
        http_response = HttpResponse(
            revision.file_object.file,
            content_type="application/zip; charset=x-user-defined",
            status=200,
        )
        http_response["Content-Disposition"] = (
            "attachment; filename=some_name.zip"
        )
    else:
        http_response = HttpResponse(status=404)
    return http_response


@login_required
@ajax_required
@require_POST
def delete_revision(request):
    response = {}
    status = 405
    revision_id = request.POST["id"]
    revision = DocumentRevision.objects.filter(pk=int(revision_id)).first()
    if revision:
        document = revision.document
        if document.owner == request.user:
            status = 200
            revision.delete()
    return JsonResponse(response, status=status)


# Check doc access rights.
def has_doc_access(doc, user):
    if doc.owner == user:
        return True
    access_rights = AccessRight.objects.filter(document=doc, user=user).first()
    if access_rights:
        return True
    else:
        return False


@login_required
@ajax_required
@require_POST
def comment_notify(request):
    response = {}
    doc_id = request.POST["doc_id"]
    collaborator_id = request.POST["collaborator_id"]
    comment_text = request.POST["comment_text"]
    comment_html = bleach.clean(request.POST["comment_html"], strip=True)
    notification_type = request.POST["type"]
    User = get_user_model()
    collaborator = User.objects.filter(pk=collaborator_id).first()
    document = Document.objects.filter(pk=doc_id).first()
    if (
        not document
        or not collaborator
        or not comment_text
        or not comment_html
        or not has_doc_access(document, request.user)
        or not notification_type
    ):
        return JsonResponse(response, status=403)
    if not has_doc_access(document, collaborator):
        # Tagged user has no access to document and will therefore not be
        # notified
        return JsonResponse(response, status=200)
    commentator = request.user.readable_name
    collaborator_name = collaborator.readable_name
    collaborator_email = collaborator.email
    document_title = document.title
    if len(document_title) == 0:
        document_title = _("Untitled")
    link = HttpRequest.build_absolute_uri(request, document.get_absolute_url())
    emails.send_comment_notification(
        notification_type,
        commentator,
        collaborator_name,
        collaborator_email,
        link,
        document_title,
        comment_text,
        comment_html,
    )
    return JsonResponse(response, status=200)


@login_required
@ajax_required
@require_POST
def get_template_for_doc(request):
    doc_id = request.POST.get("id")
    doc = (
        Document.objects.filter(id=doc_id)
        .filter(Q(owner=request.user) | Q(accessright__user=request.user))
        .select_related("template")
        .prefetch_related(
            "template__exporttemplate_set",
            "template__documentstyle_set",
            "template__documentstyle_set__documentstylefile_set",
        )
        .first()
    )
    if doc is None:
        return JsonResponse({}, status=401)
    doc_template = doc.template
    serializer = PythonWithURLSerializer()
    export_templates = serializer.serialize(
        doc_template.exporttemplate_set.all()
    )
    document_styles = serializer.serialize(
        doc_template.documentstyle_set.all(),
        use_natural_foreign_keys=True,
        fields=["title", "slug", "contents", "documentstylefile_set"],
    )
    response = {
        "id": doc_template.id,
        "title": doc_template.title,
        "content": doc_template.content,
        "doc_version": doc_template.doc_version,
        "export_templates": export_templates,
        "document_styles": document_styles,
    }
    return JsonResponse(response, status=200)


@login_required
@ajax_required
@require_POST
def get_template(request):
    response = {}
    import_id = request.POST["import_id"]
    doc_template = DocumentTemplate.objects.filter(
        Q(user=request.user) | Q(user=None), import_id=import_id
    ).first()
    if not doc_template:
        return JsonResponse(response, status=405)
    response["template"] = {
        "id": doc_template.id,
        "title": doc_template.title,
        "content": doc_template.content,
    }
    return JsonResponse(response, status=201)


@login_required
@ajax_required
@require_POST
def get_ws_base(request):
    response = {}
    doc_id = int(request.POST.get("id"))
    if len(settings.PORTS) < 2:
        ws_url_base = "/ws"
    else:
        conn = settings.PORTS[doc_id % len(settings.PORTS)]
        ws_url_base = get_url_base(request.headers["Origin"], conn)
    response["ws_base"] = ws_url_base
    return JsonResponse(response, status=200)


# maintenance views
@staff_member_required
@ajax_required
@require_POST
def get_all_old_docs(request):
    response = {}
    status = 200
    doc_list = Document.objects.filter(
        doc_version__lt=str(FW_DOCUMENT_VERSION)
    )[:10]
    response["docs"] = serializers.serialize("json", doc_list)
    return JsonResponse(response, status=status)


@staff_member_required
@ajax_required
@require_POST
def save_doc(request):
    response = {}
    status = 200
    doc_id = request.POST["id"]
    doc = Document.objects.get(pk=int(doc_id))
    # Only looking at fields that may have changed.
    content = request.POST.get("content", False)
    bibliography = request.POST.get("bibliography", False)
    comments = request.POST.get("comments", False)
    diffs = request.POST.get("diffs", False)
    version = request.POST.get("version", False)
    if content:
        doc.content = json.loads(content)
    if bibliography:
        doc.bibliography = json.loads(bibliography)
    if comments:
        doc.comments = json.loads(comments)
    if version:
        doc.version = version
    if diffs:
        doc.diffs = json.loads(diffs)
    doc.doc_version = FW_DOCUMENT_VERSION
    doc.save()
    return JsonResponse(response, status=status)


@staff_member_required
@ajax_required
@require_POST
def get_user_biblist(request):
    response = {}
    status = 200
    user_id = request.POST["user_id"]
    response["bibList"] = serializer.serialize(
        Entry.objects.filter(entry_owner_id=user_id),
        fields=("entry_key", "entry_owner", "bib_type", "fields"),
    )
    return JsonResponse(response, status=status)


@staff_member_required
@ajax_required
@require_POST
def get_all_template_ids(request):
    response = {}
    status = 200
    templates = DocumentTemplate.objects.filter(
        doc_version__lt=str(FW_DOCUMENT_VERSION)
    ).only("id")
    response["template_ids"] = []
    for template in templates:
        response["template_ids"].append(template.id)
    return JsonResponse(response, status=status)


@staff_member_required
@ajax_required
@require_POST
def get_template_admin(request, type="all"):
    template_id = request.POST.get("id")
    doc_template = DocumentTemplate.objects.filter(pk=int(template_id)).first()
    if doc_template is None:
        return JsonResponse({}, status=405)
    response = {}
    if type in ["all", "base"]:
        response["id"] = doc_template.id
        response["title"] = doc_template.title
        response["content"] = doc_template.content
        response["doc_version"] = doc_template.doc_version
    if type in ["all", "extras"]:
        serializer = PythonWithURLSerializer()
        export_templates = serializer.serialize(
            doc_template.exporttemplate_set.all()
        )
        document_styles = serializer.serialize(
            doc_template.documentstyle_set.all(),
            use_natural_foreign_keys=True,
            fields=["title", "slug", "contents", "documentstylefile_set"],
        )
        response["export_templates"] = export_templates
        response["document_styles"] = document_styles
    return JsonResponse(response, status=200)


@staff_member_required
@ajax_required
@require_POST
def save_template(request):
    response = {}
    status = 405
    template_id = request.POST["id"]
    template = DocumentTemplate.objects.filter(pk=int(template_id)).first()
    if template:
        status = 200
        # Only looking at fields that may have changed.
        content = request.POST.get("content", False)
        if content:
            template.content = json.loads(content)
        template.doc_version = FW_DOCUMENT_VERSION
        template.save()
    return JsonResponse(response, status=status)


@staff_member_required
@ajax_required
@require_POST
def create_template_admin(request):
    response = {}
    title = request.POST.get("title")
    content = json.loads(request.POST.get("content"))
    import_id = request.POST.get("import_id")
    document_styles = json.loads(request.POST.get("document_styles"))
    export_templates = json.loads(request.POST.get("export_templates"))
    template = DocumentTemplate.objects.create(
        title=title,
        content=content,
        doc_version=FW_DOCUMENT_VERSION,
        import_id=import_id,
    )
    response["id"] = template.id
    response["title"] = template.title
    date_format = "%Y-%m-%d"
    response["added"] = template.added.strftime(date_format)
    response["updated"] = template.updated.strftime(date_format)
    files = request.FILES.getlist("files[]")
    for style in document_styles:
        doc_style = DocumentStyle.objects.create(
            title=style["title"],
            slug=style["slug"],
            contents=style["contents"],
            document_template=template,
        )
        for filepath in style["files"]:
            filename = filepath.split("/").pop()
            file = next((x for x in files if x.name == filename), None)
            if file:
                DocumentStyleFile.objects.create(file=file, style=doc_style)
    for e_template in export_templates:
        filename = e_template["file"].split("/").pop()
        file = next((x for x in files if x.name == filename), None)
        if file:
            ExportTemplate.objects.create(
                document_template=template,
                template_file=file,
                file_type=e_template["file_type"],
            )
    return JsonResponse(response, status=201)


@staff_member_required
@ajax_required
@require_POST
def get_all_revision_ids(request):
    response = {}
    status = 200
    revisions = DocumentRevision.objects.filter(
        doc_version__lt=str(FW_DOCUMENT_VERSION)
    ).only("id")
    response["revision_ids"] = []
    for revision in revisions:
        response["revision_ids"].append(revision.id)
    return JsonResponse(response, status=status)


@staff_member_required
@ajax_required
@require_POST
def update_revision(request):
    response = {}
    status = 405
    revision_id = request.POST["id"]
    revision = DocumentRevision.objects.filter(pk=int(revision_id)).first()
    if revision:
        status = 200
        # keep the filename
        file_name = revision.file_object.name.split("/")[-1]
        # Delete the FieldFile as otherwise the file remains.
        revision.file_object.delete()
        revision.file_object = request.FILES["file"]
        revision.file_object.name = file_name
        revision.doc_version = FW_DOCUMENT_VERSION
        revision.save()
    return JsonResponse(response, status=status)


@staff_member_required
@ajax_required
@require_POST
def add_images_to_doc(request):
    response = {}
    status = 201
    doc_id = request.POST["doc_id"]
    doc = Document.objects.get(id=doc_id)
    # Delete all existing image links
    DocumentImage.objects.filter(document_id=doc_id).delete()
    ids = request.POST.getlist("ids[]")
    for id in ids:
        doc_image_data = {"document": doc, "title": "Deleted"}
        image = Image.objects.filter(id=id).first()
        if image:
            user_image = image.userimage_set.all().first()
            if user_image:
                doc_image_data["title"] = user_image.title
                doc_image_data["copyright"] = user_image.copyright
        else:
            image = Image()
            image.pk = id
            image.uploader = doc.owner
            f = open(
                os.path.join(
                    settings.PROJECT_PATH, "base/static/img/error.avif"
                )
            )
            image.image.save("error.avif", File(f))
            image.save()
        doc_image_data["image"] = image
        DocumentImage.objects.create(**doc_image_data)
    return JsonResponse(response, status=status)



import time
import os

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from channels.testing import ChannelsLiveServerTestCase
from testing.selenium_helper import SeleniumHelper
from selenium.common.exceptions import StaleElementReferenceException

from testing.mail import get_outbox, empty_outbox, delete_outbox
from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import override_settings

MAIL_STORAGE_NAME = "user_profile"


@override_settings(MAIL_STORAGE_NAME=MAIL_STORAGE_NAME)
@override_settings(EMAIL_BACKEND="testing.mail.EmailBackend")
class ProfileTest(SeleniumHelper, ChannelsLiveServerTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        driver_data = cls.get_drivers(1)
        cls.driver = driver_data["drivers"][0]
        cls.client = driver_data["clients"][0]
        cls.driver.implicitly_wait(driver_data["wait_time"])
        cls.wait_time = driver_data["wait_time"]

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()
        delete_outbox(MAIL_STORAGE_NAME)
        super().tearDownClass()

    def setUp(self):
        self.base_url = self.live_server_url
        self.verificationErrors = []
        self.accept_next_alert = True
        self.user = self.create_user(
            username="Yeti", email="yeti@snowman.com", passtext="otter1"
        )
        self.login_user(self.user, self.driver, self.client)

    def tearDown(self):
        empty_outbox(MAIL_STORAGE_NAME)
        self.assertEqual([], self.verificationErrors)
        return super().tearDown()

    def assertInfoAlert(self, message):
        i = 0
        message_found = False
        while i < 100:
            i = i + 1
            try:
                if (
                    self.driver.find_element(
                        By.CSS_SELECTOR,
                        "body #alerts-outer-wrapper .alerts-info",
                    ).text
                    == message
                ):
                    message_found = True
                    break
                else:
                    time.sleep(0.1)
                    continue
            except StaleElementReferenceException:
                time.sleep(0.1)
                continue
        self.assertTrue(message_found)

    def test_edit_profile(self):
        User = get_user_model()
        driver = self.driver
        driver.get(self.base_url + "/")
        driver.find_element(By.ID, "preferences-btn").click()
        driver.find_element(By.CSS_SELECTOR, ".fw-avatar-card").click()
        driver.find_element(By.ID, "first_name").clear()
        driver.find_element(By.ID, "first_name").send_keys("Snowman")
        driver.find_element(By.ID, "last_name").clear()
        driver.find_element(By.ID, "last_name").send_keys("Yeti")
        driver.find_element(By.ID, "submit-profile").click()
        pwd_button = driver.find_element(By.ID, "fw-edit-profile-pwd")
        WebDriverWait(self.driver, self.wait_time).until(
            EC.staleness_of(pwd_button)
        )
        driver.find_element(By.ID, "fw-edit-profile-pwd").click()
        driver.find_element(By.ID, "old-password-input").clear()
        driver.find_element(By.ID, "old-password-input").send_keys("otter1")
        driver.find_element(By.ID, "new-password-input1").clear()
        driver.find_element(By.ID, "new-password-input1").send_keys("otter2")
        driver.find_element(By.ID, "new-password-input2").clear()
        driver.find_element(By.ID, "new-password-input2").send_keys("otter2")
        driver.find_element(By.XPATH, "(//button[@type='button'])[2]").click()
        self.assertInfoAlert("The password has been changed.")
        driver.find_element(By.ID, "fw-edit-profile-pwd").click()
        driver.find_element(By.ID, "old-password-input").clear()
        driver.find_element(By.ID, "old-password-input").send_keys("otter2")
        driver.find_element(By.ID, "new-password-input1").clear()
        driver.find_element(By.ID, "new-password-input1").send_keys("otter1")
        driver.find_element(By.ID, "new-password-input2").clear()
        driver.find_element(By.ID, "new-password-input2").send_keys("otter1")
        driver.find_element(By.XPATH, "(//button[@type='button'])[2]").click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.invisibility_of_element_located(
                (By.CSS_SELECTOR, "#fw-change-pwd-dialog")
            )
        )
        driver.refresh()
        try:
            self.assertEqual(
                "Yeti",
                driver.find_element(By.ID, "last_name").get_attribute("value"),
            )
        except AssertionError as e:
            self.verificationErrors.append(str(e))
        try:
            self.assertEqual(
                "Snowman",
                driver.find_element(By.ID, "first_name").get_attribute(
                    "value"
                ),
            )
        except AssertionError as e:
            self.verificationErrors.append(str(e))
        try:
            self.assertEqual(
                "Yeti",
                driver.find_element(By.ID, "username").get_attribute("value"),
            )
        except AssertionError as e:
            self.verificationErrors.append(str(e))

        # Test avatar upload
        self.assertEqual(
            len(driver.find_elements(By.CSS_SELECTOR, "#profile-avatar img")),
            0,
        )
        driver.find_element(By.ID, "edit-avatar-btn").click()
        driver.find_element(
            By.XPATH, '//*[normalize-space()="Change picture"]'
        ).click()

        image_path = os.path.join(
            settings.PROJECT_PATH, "document/tests/uploads/image.png"
        )
        driver.find_element(
            By.CSS_SELECTOR, ".ui-dialog input[type=file]"
        ).send_keys(image_path)
        driver.find_element(
            By.XPATH, '//*[normalize-space()="Upload"]'
        ).click()
        self.assertEqual(
            len(driver.find_elements(By.CSS_SELECTOR, "#profile-avatar img")),
            1,
        )
        driver.find_element(By.ID, "edit-avatar-btn").click()
        driver.find_element(
            By.XPATH, '//*[normalize-space()="Delete picture"]'
        ).click()
        driver.find_element(
            By.XPATH, '//*[normalize-space()="Delete"]'
        ).click()
        time.sleep(1)
        self.assertEqual(
            len(driver.find_elements(By.CSS_SELECTOR, "#profile-avatar img")),
            0,
        )

        # Test emails
        driver.find_element(By.ID, "add-profile-email").click()
        driver.find_element(By.ID, "new-profile-email").send_keys(
            "yeti@snowman2.com"
        )
        driver.find_element(By.CSS_SELECTOR, "button.fw-dark").click()
        self.assertInfoAlert("Confirmation e-mail sent to: yeti@snowman2.com")
        assert (
            self.driver.find_element(
                By.CSS_SELECTOR,
                ".profile-email-table tbody tr:nth-child(2) .emailaddress",
            ).text
            == "yeti@snowman2.com"
        )
        time.sleep(2)
        outbox = get_outbox(MAIL_STORAGE_NAME)
        self.assertEqual(1, len(outbox))
        # We check that yeti@snowman2.com is not verified and does not have a
        # radio button for primary email account
        self.assertEqual(
            len(
                self.driver.find_elements(
                    By.CSS_SELECTOR,
                    (
                        ".profile-email-table tbody tr:nth-child(2) .fa-check, "
                        ".profile-email-table tbody tr:nth-child(2) "
                        ".primary-email-radio"
                    ),
                )
            ),
            0,
        )
        urls = self.find_urls(outbox[0].body)
        self.driver.get(urls[0])
        login_title = self.driver.find_element(
            By.CSS_SELECTOR, ".fw-login-title"
        ).text.upper()
        self.assertEqual(login_title, "CONFIRM E-MAIL ADDRESS")
        self.driver.find_element(By.ID, "submit").click()
        self.assertEqual(
            len(
                self.driver.find_elements(
                    By.CSS_SELECTOR,
                    (
                        ".profile-email-table tbody tr:nth-child(2) .fa-check, "
                        ".profile-email-table tbody tr:nth-child(2) "
                        ".primary-email-radio"
                    ),
                )
            ),
            2,
        )
        self.driver.find_element(
            By.CSS_SELECTOR,
            ".profile-email-table tbody tr:nth-child(2) .primary-email-radio",
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR, ".ui-dialog-buttonset .fw-dark"
        ).click()
        self.assertInfoAlert("The primary email has been updated.")
        self.driver.find_element(
            By.CSS_SELECTOR,
            ".profile-email-table tbody tr:nth-child(1) .delete-email",
        ).click()
        self.driver.find_element(By.CSS_SELECTOR, "button.fw-dark").click()
        self.assertInfoAlert("Email successfully deleted!")
        self.assertEqual(
            len(
                self.driver.find_elements(
                    By.CSS_SELECTOR, ".profile-email-table tbody tr"
                )
            ),
            2,
        )
        driver.find_element(By.ID, "add-profile-email").click()
        driver.find_element(By.ID, "new-profile-email").send_keys(
            "yeti@snowman3.com"
        )
        driver.find_element(By.CSS_SELECTOR, "button.fw-dark").click()
        self.assertInfoAlert("Confirmation e-mail sent to: yeti@snowman3.com")
        assert (
            self.driver.find_element(
                By.CSS_SELECTOR,
                ".profile-email-table tbody tr:nth-child(2) .emailaddress",
            ).text
            == "yeti@snowman3.com"
        )
        time.sleep(2)
        outbox = get_outbox(MAIL_STORAGE_NAME)
        self.assertEqual(2, len(outbox))
        driver.find_element(By.ID, "add-profile-email").click()
        driver.find_element(By.ID, "new-profile-email").send_keys(
            "yeti@snowman4.com"
        )
        driver.find_element(By.CSS_SELECTOR, "button.fw-dark").click()
        self.assertInfoAlert("Confirmation e-mail sent to: yeti@snowman4.com")
        assert (
            self.driver.find_element(
                By.CSS_SELECTOR,
                ".profile-email-table tbody tr:nth-child(3) .emailaddress",
            ).text
            == "yeti@snowman4.com"
        )
        time.sleep(2)
        outbox = get_outbox(MAIL_STORAGE_NAME)
        self.assertEqual(3, len(outbox))
        # This time we log out first. We should then be redirected to the page
        # that tells us to log in after verification.
        self.driver.find_element(By.ID, "preferences-btn").click()
        self.driver.find_element(
            By.XPATH, '//*[normalize-space()="Log out"]'
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.title_is("Login - Fidus Writer")
        )
        urls = self.find_urls(outbox[1].body)
        self.driver.get(urls[0])
        login_title = self.driver.find_element(
            By.CSS_SELECTOR, ".fw-login-title"
        ).text.upper()
        self.assertEqual(login_title, "CONFIRM E-MAIL ADDRESS")
        self.driver.find_element(By.ID, "submit").click()
        time.sleep(1)
        self.assertEqual(
            self.driver.find_element(
                By.CSS_SELECTOR, ".fw-contents.prelogin h1"
            ).text,
            "Thanks for verifying!",
        )
        # This time we log in as a different user and then try to verify the
        # email. This should log us out automatically.
        user2 = self.create_user(
            username="Yeti2", email="yeti@snowman5.com", passtext="otter1"
        )
        self.login_user(user2, self.driver, self.client)
        driver.get(self.base_url + "/")
        urls = self.find_urls(outbox[2].body)
        self.driver.get(urls[0])
        login_title = self.driver.find_element(
            By.CSS_SELECTOR, ".fw-login-title"
        ).text.upper()
        self.assertEqual(login_title, "CONFIRM E-MAIL ADDRESS")
        self.driver.find_element(By.ID, "submit").click()
        time.sleep(1)
        assert (
            self.driver.find_element(
                By.CSS_SELECTOR, ".fw-contents.prelogin h1"
            ).text
            == "Thanks for verifying!"
        )
        # Log in and delete account
        self.login_user(user2, self.driver, self.client)
        self.driver.get(self.base_url + "/")
        driver.find_element(By.ID, "preferences-btn").click()
        driver.find_element(By.CSS_SELECTOR, ".fw-avatar-card").click()
        driver.find_element(By.ID, "delete-account").click()
        driver.find_element(By.ID, "username-confirmation").send_keys("Yeti2")
        driver.find_element(By.ID, "password").send_keys("otter1")
        self.assertEqual(len(User.objects.filter(username="Yeti2")), 1)
        driver.find_element(
            By.XPATH, '//*[normalize-space()="Delete"]'
        ).click()
        time.sleep(1)
        login_header = self.driver.find_elements(
            By.CSS_SELECTOR, "h1.fw-login-title"
        )
        self.assertEqual(len(login_header), 1)
        self.assertEqual(len(User.objects.filter(username="Yeti2")), 0)

    def is_element_present(self, how, what):
        try:
            self.driver.find_element(by=how, value=what)
        except NoSuchElementException:
            return False
        return True

    def test_language_preference(self):
        User = get_user_model()
        driver = self.driver
        driver.get(self.base_url + "/")

        # Go to profile page
        driver.find_element(By.ID, "preferences-btn").click()
        driver.find_element(By.CSS_SELECTOR, ".fw-avatar-card").click()

        # Find the language dropdown
        language_dropdown = driver.find_element(By.ID, "language")

        # Check if dropdown exists and has options
        self.assertTrue(self.is_element_present(By.ID, "language"))

        # Initially language should be default (empty)
        self.assertEqual(
            language_dropdown.get_attribute("value"),
            "",
            "Initial language should be default (empty)",
        )

        # Change language to German
        select = driver.find_element(By.ID, "language")
        select.find_element(By.CSS_SELECTOR, "option[value='de']").click()

        # Save the profile
        driver.find_element(By.ID, "submit-profile").click()

        time.sleep(1)

        # Reload the page to see if the language was saved
        driver.refresh()

        # Wait for page to load
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((By.ID, "language"))
        )

        # Check if the language is now German
        select = driver.find_element(By.ID, "language")
        self.assertEqual(
            select.get_attribute("value"),
            "de",
            "Language should be set to German",
        )

        # Change language back to default
        select.find_element(By.CSS_SELECTOR, "option[value='']").click()

        # Save the profile
        driver.find_element(By.ID, "submit-profile").click()

        time.sleep(1)

        # Check if the user in the database has the updated language
        user = User.objects.get(username="Yeti")
        self.assertEqual(
            user.language, None, "Language should be reset to default (None)"
        )

        # Set language to Spanish
        driver.refresh()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((By.ID, "language"))
        )
        select = driver.find_element(By.ID, "language")
        select.find_element(By.CSS_SELECTOR, "option[value='es']").click()
        driver.find_element(By.ID, "submit-profile").click()

        # Wait for the save to complete
        time.sleep(1)

        # Log out to test if language persists on login
        driver.find_element(By.ID, "preferences-btn").click()

        driver.find_element(
            By.XPATH, '//*[normalize-space()="Cerrar sesión"]'
        ).click()

        # Wait for login page
        WebDriverWait(self.driver, self.wait_time).until(
            EC.title_is("Login - Fidus Writer")
        )

        # Login again
        driver.find_element(By.ID, "id-login").send_keys("Yeti")
        driver.find_element(By.ID, "id-password").send_keys("otter1")
        driver.find_element(By.ID, "login-submit").click()

        # Wait for dashboard page
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((By.ID, "preferences-btn"))
        )

        # Go to profile page
        driver.find_element(By.ID, "preferences-btn").click()
        driver.find_element(By.CSS_SELECTOR, ".fw-avatar-card").click()

        # Check if the language is still Spanish
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((By.ID, "language"))
        )
        select = driver.find_element(By.ID, "language")
        self.assertEqual(
            select.get_attribute("value"),
            "es",
            "Language should still be set to Spanish after logout/login",
        )

        # Check if the user in the database has the updated language
        user = User.objects.get(username="Yeti")
        self.assertEqual(
            user.language,
            "es",
            "Language should be set to Spanish in database",
        )



# Generated by Django 5.1.7 on 2025-04-08 13:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("user", "0011_delete_loginas_alter_user_options"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="language",
            field=models.CharField(
                blank=True,
                choices=[
                    ("en", "English"),
                    ("bg", "Bulgarian"),
                    ("de", "German"),
                    ("fr", "French"),
                    ("it", "Italian"),
                    ("es", "Spanish"),
                    ("pt-br", "Portuguese (Brazil)"),
                ],
                default=None,
                help_text="Interface language preference",
                max_length=10,
                null=True,
            ),
        ),
    ]



import autobahn
import uuid
import atexit
import logging
import gc
import asyncio
from time import mktime, time
from copy import deepcopy

from asgiref.sync import sync_to_async
from django.db.utils import DatabaseError
from django.db.models import F, Q
from django.conf import settings
from django.db.utils import IntegrityError

from base.helpers.ws import get_url_base
from document.helpers.session_user_info import SessionUserInfo
from document import prosemirror
from document.helpers.serializers import PythonWithURLSerializer
from base.base_consumer import BaseWebsocketConsumer
from document.models import (
    COMMENT_ONLY,
    CAN_UPDATE_DOCUMENT,
    CAN_COMMUNICATE,
    FW_DOCUMENT_VERSION,
    DocumentTemplate,
)
from usermedia.models import Image, DocumentImage, UserImage
from user.helpers import Avatars

logger = logging.getLogger(__name__)


class WebsocketConsumer(BaseWebsocketConsumer):
    sessions = dict()
    history_length = 1000  # Only keep the last 1000 diffs

    async def connect(self):
        self.document_id = int(
            self.scope["url_route"]["kwargs"]["document_id"]
        )
        redirected = await self.check_server()
        if redirected:
            return
        connected = await super().connect()
        if not connected:
            return
        logger.debug(
            f"Action:Document socket opened by user. "
            f"URL:{self.endpoint} User:{self.user.id} ParticipantID:{self.id}"
        )

    async def check_server(self):
        # The correct server for the document may have changed since the client
        # received its initial connection information. For example, because the
        # number of servers has increased (all servers need to restart to have
        # the right setting).
        if len(settings.PORTS) < 2:
            return False
        # We compare the internal port
        actual_port = self.scope["server"][1]
        expected_conn = settings.PORTS[self.document_id % len(settings.PORTS)]
        expected_port = (
            expected_conn["internal"]
            if isinstance(expected_conn, dict)
            else expected_conn
        )
        if actual_port != expected_port:
            # Redirect to the correct URL
            await self.init()
            origin = (
                dict(self.scope["headers"]).get(b"origin", b"").decode("utf-8")
            )
            expected = get_url_base(origin, expected_conn)
            logger.debug(f"Redirecting from {actual_port} to {expected}.")
            await self.send_message({"type": "redirect", "base": expected})
            await self.do_close()
            return True
        return False

    async def confirm_diff(self, rid):
        response = {"type": "confirm_diff", "rid": rid}
        await self.send_message(response)

    async def subscribe(self, connection_count=0):
        # Create a new instance
        self.user_info = SessionUserInfo(self.user)

        # Initialize access asynchronously
        doc_db, can_access = await self.user_info.init_access(self.document_id)

        if not can_access or float(doc_db.doc_version) != FW_DOCUMENT_VERSION:
            await self.access_denied()
            return

        if (
            doc_db.id in WebsocketConsumer.sessions
            and len(WebsocketConsumer.sessions[doc_db.id]["participants"]) > 0
        ):
            logger.debug(
                f"Action:Serving already opened document. "
                f"URL:{self.endpoint} User:{self.user.id} "
                f" ParticipantID:{self.id}"
            )
            self.session = WebsocketConsumer.sessions[doc_db.id]
            self.id = max(self.session["participants"]) + 1
            self.session["participants"][self.id] = self
            template = False
        else:
            logger.debug(
                f"Action:Opening document from DB. "
                f"URL:{self.endpoint} User:{self.user.id} "
                f"ParticipantID:{self.id}"
            )
            self.id = 0
            if "type" not in doc_db.content:
                doc_db.content = deepcopy(doc_db.template.content)
                if "type" not in doc_db.content:
                    doc_db.content["type"] = "doc"
                if "content" not in doc_db.content:
                    doc_db.content["content"] = [{type: "title"}]
                await doc_db.asave()
            node = prosemirror.from_json(doc_db.content)
            self.session = {
                "doc": doc_db,
                "node": node,
                "node_updates": False,
                "participants": {0: self},
                "last_saved_version": doc_db.version,
            }
            WebsocketConsumer.sessions[doc_db.id] = self.session
            if self.user_info.access_rights == "write":
                template = True
            else:
                template = False
        logger.debug(
            f"Action:Participant ID Assigned. URL:{self.endpoint} "
            f"User:{self.user.id} ParticipantID:{self.id}"
        )
        await self.send_message({"type": "subscribed"})
        if connection_count < 1:
            await self.send_styles()
            await self.send_document(False, template)
        if connection_count >= 1:
            # If the user is reconnecting pass along access_rights to
            # front end to compare it with previous rights.
            await self.send_message(
                {
                    "type": "access_right",
                    "access_right": self.user_info.access_rights,
                }
            )
        if await self.can_communicate():
            await self.handle_participant_update()

    async def send_styles(self):
        doc_db = self.session["doc"]
        response = dict()
        response["type"] = "styles"

        # Create serializer and get data
        serializer = PythonWithURLSerializer()

        # These operations could be expensive - run them in parallel
        export_temps_task = sync_to_async(serializer.serialize)(
            doc_db.template.exporttemplate_set.all(),
            fields=["file_type", "template_file", "title"],
        )

        document_styles_task = sync_to_async(serializer.serialize)(
            doc_db.template.documentstyle_set.all(),
            use_natural_foreign_keys=True,
            fields=["title", "slug", "contents", "documentstylefile_set"],
        )

        # Run queries in parallel using asyncio.gather
        export_temps, document_styles = await asyncio.gather(
            export_temps_task, document_styles_task
        )

        # Get document templates
        document_templates = {}
        query = DocumentTemplate.objects.filter(
            Q(user=self.user) | Q(user=None)
        ).order_by(F("user").desc(nulls_first=True))

        # Use async for to properly consume the async iterator
        async for obj in query.aiterator():
            document_templates[obj.import_id] = {
                "title": obj.title,
                "id": obj.id,
            }

        response["styles"] = {
            "export_templates": [obj["fields"] for obj in export_temps],
            "document_styles": [obj["fields"] for obj in document_styles],
            "document_templates": document_templates,
        }
        await self.send_message(response)

    async def unfixable(self):
        await self.send_document()

    async def send_document(self, messages=False, template=False):
        response = dict()
        response["type"] = "doc_data"
        doc_owner = self.session["doc"].owner
        avatars = Avatars()

        # Use async avatar URL generation
        owner_avatar = await avatars.get_url_async(doc_owner)

        response["doc_info"] = {
            "id": self.session["doc"].id,
            "is_owner": self.user_info.is_owner,
            "access_rights": self.user_info.access_rights,
            "path": self.user_info.path,
            "owner": {
                "id": doc_owner.id,
                "name": doc_owner.readable_name,
                "username": doc_owner.username,
                "avatar": owner_avatar,
                "contacts": [],
            },
        }
        await sync_to_async(WebsocketConsumer.serialize_content)(self.session)
        response["doc"] = {
            "v": self.session["doc"].version,
            "content": self.session["doc"].content,
            "bibliography": self.session["doc"].bibliography,
            "images": {},
        }
        if template:
            response["doc"]["template"] = {
                "id": self.session["doc"].template.id,
                "content": self.session["doc"].template.content,
            }
        if messages:
            response["m"] = messages
        response["time"] = int(time()) * 1000

        # Get document images
        doc_id = self.session["doc"].id
        async for dimage in (
            DocumentImage.objects.filter(document_id=doc_id)
            .select_related("image")
            .only(
                "id",
                "title",
                "copyright",
                "image__id",
                "image__image",
                "image__file_type",
                "image__added",
                "image__checksum",
                "image__thumbnail",
                "image__height",
                "image__width",
            )
            .aiterator()
        ):
            image = dimage.image
            field_obj = {
                "id": image.id,
                "title": dimage.title,
                "copyright": dimage.copyright,
                "image": image.image.url,
                "file_type": image.file_type,
                "added": mktime(image.added.timetuple()) * 1000,
                "checksum": image.checksum,
                "cats": [],
            }
            if image.thumbnail:
                field_obj["thumbnail"] = image.thumbnail.url
                field_obj["height"] = image.height
                field_obj["width"] = image.width
            response["doc"]["images"][image.id] = field_obj

        if self.user_info.access_rights == "read-without-comments":
            response["doc"]["comments"] = []
        elif self.user_info.access_rights in ["review", "review-tracked"]:
            # Reviewer should only get his/her own comments
            filtered_comments = {}
            for key, value in list(self.session["doc"].comments.items()):
                if value["user"] == self.user_info.user.id:
                    filtered_comments[key] = value
            response["doc"]["comments"] = filtered_comments
        else:
            response["doc"]["comments"] = self.session["doc"].comments

        # Get contacts asynchronously
        contacts = []
        async for contact in (
            doc_owner.contacts.all()
            .only("id", "username", "first_name", "last_name")
            .aiterator()
        ):
            contacts.append(contact)

        # Get all avatars in parallel
        contact_avatars = await asyncio.gather(
            *[avatars.get_url_async(contact) for contact in contacts]
        )

        # Now add contacts with their avatars
        for i, contact in enumerate(contacts):
            contact_object = {
                "id": contact.id,
                "name": contact.readable_name,
                "username": contact.get_username(),
                "avatar": contact_avatars[i],
                "type": "user",
            }
            response["doc_info"]["owner"]["contacts"].append(contact_object)

        if self.user_info.is_owner:
            invites = []
            async for invite in doc_owner.invites_by.all().aiterator():
                invites.append(invite)
            for contact in invites:
                contact_object = {
                    "id": contact.id,
                    "name": contact.username,
                    "username": contact.username,
                    "avatar": None,
                    "type": "userinvite",
                }
                response["doc_info"]["owner"]["contacts"].append(
                    contact_object
                )

        response["doc_info"]["session_id"] = self.id
        await self.send_message(response)

    async def reject_message(self, message):
        if message["type"] == "diff":
            await self.send_message(
                {"type": "reject_diff", "rid": message["rid"]}
            )

    async def handle_message(self, message):
        if self.user_info.document_id not in WebsocketConsumer.sessions:
            logger.debug(
                f"Action:Receiving message for closed document. "
                f"URL:{self.endpoint} User:{self.user.id} "
                f"ParticipantID:{self.id}"
            )
            return
        if message["type"] == "get_document":
            await self.send_document()
        elif (
            message["type"] == "participant_update"
            and await self.can_communicate()
        ):
            await self.handle_participant_update()
        elif message["type"] == "chat" and await self.can_communicate():
            await self.handle_chat(message)
        elif message["type"] == "check_version":
            await self.check_version(message)
        elif message["type"] == "selection_change":
            await self.handle_selection_change(message)
        elif message["type"] == "diff" and await self.can_update_document():
            await self.handle_diff(message)
        elif message["type"] == "path_change":
            await self.handle_path_change(message)

    async def update_bibliography(self, bibliography_updates):
        for bu in bibliography_updates:
            if "id" not in bu:
                continue
            id = bu["id"]
            if bu["type"] == "update":
                self.session["doc"].bibliography[id] = bu["reference"]
            elif bu["type"] == "delete":
                del self.session["doc"].bibliography[id]

    async def update_images(self, image_updates):
        for iu in image_updates:
            if "id" not in iu:
                continue
            id = iu["id"]
            if iu["type"] == "update":
                # Ensure that access rights exist
                has_access = await UserImage.objects.filter(
                    image__id=id, owner=self.user_info.user
                ).aexists()

                if not has_access:
                    continue

                doc_image = await DocumentImage.objects.filter(
                    document_id=self.session["doc"].id, image_id=id
                ).afirst()

                if doc_image:
                    doc_image.title = iu["image"]["title"]
                    doc_image.copyright = iu["image"]["copyright"]
                    await doc_image.asave()
                else:
                    await DocumentImage.objects.acreate(
                        document_id=self.session["doc"].id,
                        image_id=id,
                        title=iu["image"]["title"],
                        copyright=iu["image"]["copyright"],
                    )
            elif iu["type"] == "delete":
                await DocumentImage.objects.filter(
                    document_id=self.session["doc"].id, image_id=id
                ).adelete()

                async for image in Image.objects.filter(id=id).aiterator():
                    can_delete = await sync_to_async(image.is_deletable)()
                    if can_delete:
                        await image.adelete()

    async def update_comments(self, comments_updates):
        comments_updates = deepcopy(comments_updates)
        for cd in comments_updates:
            if "id" not in cd:
                # ignore
                continue
            id = cd["id"]
            if cd["type"] == "create":
                self.session["doc"].comments[id] = {
                    "user": cd["user"],
                    "username": cd["username"],
                    "assignedUser": cd["assignedUser"],
                    "assignedUsername": cd["assignedUsername"],
                    "date": cd["date"],
                    "comment": cd["comment"],
                    "isMajor": cd["isMajor"],
                    "resolved": cd["resolved"],
                }
            elif cd["type"] == "delete":
                del self.session["doc"].comments[id]
            elif cd["type"] == "update":
                self.session["doc"].comments[id]["comment"] = cd["comment"]
                if "isMajor" in cd:
                    self.session["doc"].comments[id]["isMajor"] = cd["isMajor"]
                if "assignedUser" in cd and "assignedUsername" in cd:
                    self.session["doc"].comments[id]["assignedUser"] = cd[
                        "assignedUser"
                    ]
                    self.session["doc"].comments[id]["assignedUsername"] = cd[
                        "assignedUsername"
                    ]
                if "resolved" in cd:
                    self.session["doc"].comments[id]["resolved"] = cd[
                        "resolved"
                    ]
            elif cd["type"] == "add_answer":
                if "answers" not in self.session["doc"].comments[id]:
                    self.session["doc"].comments[id]["answers"] = []
                self.session["doc"].comments[id]["answers"].append(
                    {
                        "id": cd["answerId"],
                        "user": cd["user"],
                        "username": cd["username"],
                        "date": cd["date"],
                        "answer": cd["answer"],
                    }
                )
            elif cd["type"] == "delete_answer":
                answer_id = cd["answerId"]
                for answer in self.session["doc"].comments[id]["answers"]:
                    if answer["id"] == answer_id:
                        self.session["doc"].comments[id]["answers"].remove(
                            answer
                        )
            elif cd["type"] == "update_answer":
                answer_id = cd["answerId"]
                for answer in self.session["doc"].comments[id]["answers"]:
                    if answer["id"] == answer_id:
                        answer["answer"] = cd["answer"]

    async def handle_participant_update(self):
        await WebsocketConsumer.send_participant_list(
            self.user_info.document_id
        )

    async def handle_chat(self, message):
        chat = {
            "id": str(uuid.uuid4()),
            "body": message["body"],
            "from": self.user_info.user.id,
            "type": "chat",
        }
        await WebsocketConsumer.send_updates(chat, self.user_info.document_id)

    async def handle_selection_change(self, message):
        if (
            self.user_info.document_id in WebsocketConsumer.sessions
            and message["v"] == self.session["doc"].version
        ):
            await WebsocketConsumer.send_updates(
                message,
                self.user_info.document_id,
                self.id,
                self.user_info.user.id,
            )

    async def handle_path_change(self, message):
        if (
            self.user_info.document_id in WebsocketConsumer.sessions
            and self.user_info.path_object
        ):
            self.user_info.path_object.path = message["path"]
            await self.user_info.path_object.asave(
                update_fields=[
                    "path",
                ]
            )
            await WebsocketConsumer.send_updates(
                message,
                self.user_info.document_id,
                self.id,
                self.user_info.user.id,
            )

    # Checks if the diff only contains changes to comments.
    def only_comments(self, message):
        allowed_operations = ["addMark", "removeMark"]
        only_comment = True
        if "ds" in message:  # ds = document steps
            for step in message["ds"]:
                if not (
                    step["stepType"] in allowed_operations
                    and step["mark"]["type"] == "comment"
                ):
                    only_comment = False
        return only_comment

    async def handle_diff(self, message):
        pv = message["v"]
        dv = self.session["doc"].version
        logger.debug(
            f"Action:Handling Diff. URL:{self.endpoint} User:{self.user.id} "
            f"ParticipantID:{self.id} Client version:{pv} "
            f"Server version:{dv} Message:{message}"
        )
        if (
            self.user_info.access_rights in COMMENT_ONLY
            and not self.only_comments(message)
        ):
            logger.error(
                f"Action:Received non-comment diff from comment-only "
                f"collaborator.Discarding URL:{self.endpoint} "
                f"User:{self.user.id} ParticipantID:{self.id}"
            )
            return
        if pv == dv:
            if "ds" in message:  # ds = document steps
                updated_node = prosemirror.apply(
                    message["ds"], self.session["node"]
                )
                if updated_node:
                    self.session["node"] = updated_node
                    self.session["node_updates"] = True
                else:
                    await self.unfixable()
                    patch_msg = {
                        "type": "patch_error",
                        "user_id": self.user.id,
                    }
                    await self.send_message(patch_msg)
                    # Reset collaboration to avoid any data loss issues.
                    await self.reset_collaboration(
                        patch_msg, self.user_info.document_id, self.id
                    )
                    return
            self.session["doc"].diffs.append(message)
            self.session["doc"].diffs = self.session["doc"].diffs[
                -self.history_length :
            ]
            self.session["doc"].version += 1
            if "ti" in message:  # ti = title
                self.session["doc"].title = message["ti"][-255:]
            if "cu" in message:  # cu = comment updates
                await self.update_comments(message["cu"])
            if "bu" in message:  # bu = bibliography updates
                await self.update_bibliography(message["bu"])
            if "iu" in message:  # iu = image updates
                await self.update_images(message["iu"])
            if self.session["doc"].version % settings.DOC_SAVE_INTERVAL == 0:
                await WebsocketConsumer.save_document_async(
                    self.user_info.document_id
                )
            await self.confirm_diff(message["rid"])
            await WebsocketConsumer.send_updates(
                message,
                self.user_info.document_id,
                self.id,
                self.user_info.user.id,
            )
        elif pv < dv:
            if pv + len(self.session["doc"].diffs) >= dv:
                # We have enough diffs stored to fix it.
                number_diffs = dv - pv
                logger.debug(
                    f"Action:Resending document diffs. URL:{self.endpoint} "
                    f"User:{self.user.id} ParticipantID:{self.id} "
                    f"number of messages to be resent:{number_diffs}"
                )
                messages = self.session["doc"].diffs[-number_diffs:]
                for message in messages:
                    new_message = message.copy()
                    new_message["server_fix"] = True
                    await self.send_message(new_message)
            else:
                logger.debug(
                    f"Action:User is on a very old version of the document. "
                    f"URL:{self.endpoint} User:{self.user.id} "
                    f"ParticipantID:{self.id}"
                )
                # Client has a version that is too old to be fixed
                await self.unfixable()
                return
        else:
            # Client has a higher version than server. Something is fishy!
            logger.debug(
                f"Action:User has higher document version than server.Fishy! "
                f"URL:{self.endpoint} User:{self.user.id} "
                f"ParticipantID:{self.id}"
            )

    async def check_version(self, message):
        pv = message["v"]
        dv = self.session["doc"].version
        logger.debug(
            f"Action:Checking version of document. URL:{self.endpoint} "
            f"User:{self.user.id} ParticipantID:{self.id} "
            f"Client document version:{pv} Server document version:{dv}"
        )
        if pv == dv:
            response = {
                "type": "confirm_version",
                "v": pv,
            }
            await self.send_message(response)
            return
        elif pv > dv:
            logger.debug(
                f"Action:User is on a newer version of the document. "
                f"URL:{self.endpoint} User:{self.user.id} "
                f"ParticipantID:{self.id}"
            )
            await self.unfixable()
            return
        elif pv + len(self.session["doc"].diffs) >= dv:
            number_diffs = dv - pv
            logger.debug(
                f"Action:Resending document diffs. URL:{self.endpoint} "
                f"User:{self.user.id} ParticipantID:{self.id}"
                f"number of messages to be resent:{number_diffs}"
            )
            messages = self.session["doc"].diffs[-number_diffs:]
            await self.send_document(messages)
            return
        else:
            logger.debug(
                f"Action:User is on a very old version of the document. "
                f"URL:{self.endpoint} User:{self.user.id} "
                f"ParticipantID:{self.id}"
            )
            # Client has a version that is too old
            await self.unfixable()
            return

    async def can_update_document(self):
        return self.user_info.access_rights in CAN_UPDATE_DOCUMENT

    async def can_communicate(self):
        return self.user_info.access_rights in CAN_COMMUNICATE

    async def disconnect(self, code):
        if (
            hasattr(self, "endpoint")
            and hasattr(self, "user")
            and hasattr(self, "id")
        ):
            logger.debug(
                f"Action:Closing websocket. URL:{self.endpoint} "
                f"User:{self.user.id} ParticipantID:{self.id}"
            )
        if (
            hasattr(self, "session")
            and hasattr(self, "user_info")
            and hasattr(self.user_info, "document_id")
        ):
            doc_id = self.user_info.document_id
            if doc_id in WebsocketConsumer.sessions:
                # Clear this participant's specific resources
                if (
                    hasattr(self, "id")
                    and self.id in self.session["participants"]
                ):
                    # Remove this participant
                    self.session["participants"].pop(self.id)

                # Complete document cleanup if no participants remain
                if len(self.session["participants"]) == 0:
                    # Save before cleanup
                    await WebsocketConsumer.save_document_async(doc_id)

                    # Break references manually before deleting
                    session = WebsocketConsumer.sessions[doc_id]

                    # Clear prosemirror node structure
                    if "node" in session:
                        # Recursively break node references if possible
                        session["node"] = None

                    # Clear diff history completely
                    if "doc" in session and hasattr(session["doc"], "diffs"):
                        session["doc"].diffs = None  # Not just an empty list

                    # Remove complete session
                    WebsocketConsumer.sessions.pop(doc_id, None)

                    # Force garbage collection
                    gc.collect()

                else:
                    try:
                        # Update participant list if there are still participants
                        await WebsocketConsumer.send_participant_list(
                            self.user_info.document_id
                        )
                    except autobahn.exception.Disconnected:
                        logger.error(
                            "Error sending participant list over disconnected session"
                        )

            # Clear any remaining references to large objects
            if hasattr(self, "session"):
                self.session = None
            if hasattr(self, "user_info"):
                self.user_info = None
            await self.close()

    @classmethod
    async def send_participant_list(cls, document_id):
        if document_id in WebsocketConsumer.sessions:
            avatars = Avatars()
            participant_list = []

            # Create a list of tasks to get avatars
            avatar_tasks = []
            participants_data = []

            for session_id, waiter in list(
                cls.sessions[document_id]["participants"].items()
            ):
                access_rights = waiter.user_info.access_rights
                if access_rights not in CAN_COMMUNICATE:
                    continue
                participants_data.append(
                    {
                        "session_id": session_id,
                        "id": waiter.user_info.user.id,
                        "name": waiter.user_info.user.readable_name,
                        "user": waiter.user_info.user,
                    }
                )
                # Add task to get avatar
                avatar_tasks.append(
                    avatars.get_url_async(waiter.user_info.user)
                )

            # Get all avatars in parallel
            if avatar_tasks:
                avatar_urls = await asyncio.gather(*avatar_tasks)

                # Now build the participant list with avatars
                for i, data in enumerate(participants_data):
                    participant_list.append(
                        {
                            "session_id": data["session_id"],
                            "id": data["id"],
                            "name": data["name"],
                            "avatar": avatar_urls[i],
                        }
                    )

            message = {
                "participant_list": participant_list,
                "type": "connections",
            }
            await WebsocketConsumer.send_updates(message, document_id)

    @classmethod
    async def reset_collaboration(
        cls, patch_exception_msg, document_id, sender_id
    ):
        logger.debug(
            f"Action:Resetting collaboration. DocumentID:{document_id} "
            f"Patch conflict triggered. ParticipantID:{sender_id} "
            f"waiters:{len(cls.sessions[document_id]['participants'])}"
        )

        # Create a list of coroutines to execute
        tasks = []

        for waiter in list(cls.sessions[document_id]["participants"].values()):
            if waiter.id != sender_id:
                tasks.append(waiter.unfixable())
                tasks.append(waiter.send_message(patch_exception_msg))

        # Execute all tasks concurrently
        if tasks:
            await asyncio.gather(*tasks)

    @classmethod
    async def send_updates(
        cls, message, document_id, sender_id=None, user_id=None
    ):
        logger.debug(
            f"Action:Sending message to waiters. DocumentID:{document_id} "
            f"waiters:{len(cls.sessions[document_id]['participants'])}"
        )

        # Create a list of send tasks to execute concurrently
        send_tasks = []

        for waiter in list(cls.sessions[document_id]["participants"].values()):
            if waiter.id != sender_id:
                access_rights = waiter.user_info.access_rights
                msg_to_send = message

                # Check if we need to modify the message based on access rights
                need_copy = False

                if "comments" in message and len(message["comments"]) > 0:
                    # Filter comments if needed
                    if access_rights == "read-without-comments":
                        need_copy = True
                    elif (
                        access_rights in ["review", "review-tracked"]
                        and user_id != waiter.user_info.user.id
                    ):
                        need_copy = True
                elif (
                    message["type"] in ["chat", "connections"]
                    and access_rights not in CAN_COMMUNICATE
                ):
                    continue
                elif (
                    message["type"] == "selection_change"
                    and access_rights not in CAN_COMMUNICATE
                    and user_id != waiter.user_info.user.id
                ):
                    continue
                elif (
                    message["type"] == "path_change"
                    and user_id != waiter.user_info.user.id
                ):
                    continue

                # Create a copy of the message if needed to modify it
                if need_copy:
                    msg_to_send = deepcopy(message)
                    msg_to_send["comments"] = []

                # Add the send task to our list
                send_tasks.append(waiter.send_message(msg_to_send))

        # Execute all send tasks concurrently
        if send_tasks:
            await asyncio.gather(*send_tasks)

    @classmethod
    def serialize_content(cls, session):
        if "node_updates" in session and session["node_updates"]:
            session["doc"].content = prosemirror.to_mini_json(session["node"])
            session["node_updates"] = False

    @classmethod
    async def save_document_async(cls, document_id):
        session = cls.sessions[document_id]
        if session["doc"].version == session["last_saved_version"]:
            return
        logger.debug(
            f"Action:Saving document to DB. DocumentID:{session['doc'].id} "
            f"Doc version:{session['doc'].version}"
        )
        cls.serialize_content(session)
        try:
            # this try block is to avoid a db exception
            # in case the doc has been deleted from the db
            # in fiduswriter the owner of a doc could delete a doc
            # while an invited writer is editing the same doc
            await session["doc"].asave(
                update_fields=[
                    "title",
                    "version",
                    "content",
                    "diffs",
                    "comments",
                    "bibliography",
                    "updated",
                ]
            )

        except DatabaseError as e:
            expected_msg = "Save with update_fields did not affect any rows."
            if str(e) == expected_msg:
                try:
                    await session["doc"].asave()
                except IntegrityError:
                    pass
            else:
                raise e
        session["last_saved_version"] = session["doc"].version

    @classmethod
    def save_document(cls, document_id):
        session = cls.sessions[document_id]
        if session["doc"].version == session["last_saved_version"]:
            return
        logger.debug(
            f"Action:Saving document to DB. DocumentID:{session['doc'].id} "
            f"Doc version:{session['doc'].version}"
        )
        cls.serialize_content(session)
        try:
            # this try block is to avoid a db exception
            # in case the doc has been deleted from the db
            # in fiduswriter the owner of a doc could delete a doc
            # while an invited writer is editing the same doc
            session["doc"].save(
                update_fields=[
                    "title",
                    "version",
                    "content",
                    "diffs",
                    "comments",
                    "bibliography",
                    "updated",
                ]
            )

        except DatabaseError as e:
            expected_msg = "Save with update_fields did not affect any rows."
            if str(e) == expected_msg:
                try:
                    session["doc"].save()
                except IntegrityError:
                    pass
            else:
                raise e
        session["last_saved_version"] = session["doc"].version

    @classmethod
    def save_all_docs(cls):
        for document_id in cls.sessions:
            cls.save_document(document_id)


atexit.register(WebsocketConsumer.save_all_docs)



# Generated by Django 1.11.13 on 2018-08-14 17:37
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import usermedia.models


class Migration(migrations.Migration):
    replaces = [
        ("usermedia", "0001_initial"),
        ("usermedia", "0002_auto_20151226_1110"),
        ("usermedia", "0003_auto_20160218_2222"),
        ("usermedia", "0004_auto_20160218_2250"),
        ("usermedia", "0005_userimage"),
        ("usermedia", "0006_auto_20170813_2154"),
        ("usermedia", "0007_auto_20170813_2206"),
        ("usermedia", "0008_documentimage"),
        ("usermedia", "0009_auto_20170908_0953"),
    ]

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("document", "0001_squashed_20200219"),
    ]

    operations = [
        migrations.CreateModel(
            name="Image",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=128)),
                ("added", models.DateTimeField(auto_now_add=True)),
                (
                    "image",
                    models.FileField(upload_to=usermedia.models.get_file_path),
                ),
                (
                    "thumbnail",
                    models.ImageField(
                        blank=True,
                        max_length=500,
                        null=True,
                        upload_to="image_thumbnails",
                    ),
                ),
                ("image_cat", models.CharField(default="", max_length=255)),
                (
                    "file_type",
                    models.CharField(blank=True, max_length=20, null=True),
                ),
                ("height", models.IntegerField(blank=True, null=True)),
                ("width", models.IntegerField(blank=True, null=True)),
                ("checksum", models.BigIntegerField(default=0)),
                (
                    "uploader",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="uploader",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ImageCategory",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("category_title", models.CharField(max_length=100)),
                (
                    "category_owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Image categories",
            },
        ),
        migrations.AddField(
            model_name="image",
            name="owner",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="image_owner",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="image",
            name="uploader",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="image_uploader",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.CreateModel(
            name="UserImage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=128)),
                ("image_cat", models.CharField(default="", max_length=255)),
                (
                    "image",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="usermedia.Image",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="image_owner",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.RemoveField(
            model_name="image",
            name="image_cat",
        ),
        migrations.RemoveField(
            model_name="image",
            name="owner",
        ),
        migrations.RemoveField(
            model_name="image",
            name="title",
        ),
        migrations.CreateModel(
            name="DocumentImage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(default="", max_length=128)),
                (
                    "document",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="document.Document",
                    ),
                ),
                (
                    "image",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="usermedia.Image",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 2.2.9 on 2020-02-19 13:48

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import document.models
import uuid


class Migration(migrations.Migration):
    replaces = [
        ("document", "0001_squashed_0032_document_listed"),
        ("document", "0002_auto_20180814_1946"),
        ("document", "0003_auto_20181115_0926"),
        ("document", "0004_documenttemplate"),
        ("document", "0005_document_template"),
        ("document", "0006_auto_20190118_1342"),
        ("document", "0007_auto_20190227_2105"),
        ("document", "0008_documenttemplate_doc_version"),
        ("document", "0009_documentrevision_doc_version"),
        ("document", "0008_auto_20190701_0943"),
        ("document", "0010_merge_20190701_1737"),
        ("document", "0011_auto_20190701_1744"),
        ("document", "0012_auto_20190704_1219"),
        ("document", "0013_auto_20190808_1126"),
        ("document", "0014_auto_20190811_1204"),
        ("document", "0015_auto_20190828_1108"),
        ("document", "0016_auto_20190829_2147"),
        ("document", "0017_remove_documenttemplate_citation_styles"),
        ("document", "0018_documenttemplate_import_id"),
        ("document", "0019_remove_documenttemplate_definition_hash"),
    ]

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="DocumentTemplate",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(blank=True, default="", max_length=255),
                ),
                ("definition", models.TextField(default="{}")),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "doc_version",
                    models.DecimalField(
                        decimal_places=1, default=3.1, max_digits=3
                    ),
                ),
                ("added", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                (
                    "import_id",
                    models.CharField(blank=True, default="", max_length=255),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Document",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(blank=True, default="", max_length=255),
                ),
                ("contents", models.TextField(default="{}")),
                ("added", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="owner",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                ("last_diffs", models.TextField(default="[]")),
                ("version", models.PositiveIntegerField(default=0)),
                ("comments", models.TextField(default="{}")),
                (
                    "doc_version",
                    models.DecimalField(
                        decimal_places=1, default=3.1, max_digits=3
                    ),
                ),
                ("bibliography", models.TextField(default="{}")),
                ("listed", models.BooleanField(default=True)),
                (
                    "template",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="document.DocumentTemplate",
                    ),
                ),
            ],
            options={
                "ordering": ["-id"],
            },
        ),
        migrations.CreateModel(
            name="AccessRight",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "rights",
                    models.CharField(
                        choices=[
                            ("write", "Writer"),
                            ("write-tracked", "Write with tracked changes"),
                            ("comment", "Commentator"),
                            ("review", "Reviewer"),
                            ("read", "Reader"),
                            (
                                "read-without-comments",
                                "Reader without comment access",
                            ),
                        ],
                        max_length=21,
                    ),
                ),
                (
                    "document",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="document.Document",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "unique_together": {("document", "user")},
            },
        ),
        migrations.CreateModel(
            name="AccessRightInvite",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("email", models.EmailField(max_length=254)),
                (
                    "rights",
                    models.CharField(
                        choices=[
                            ("write", "Writer"),
                            ("write-tracked", "Write with tracked changes"),
                            ("comment", "Commentator"),
                            ("review", "Reviewer"),
                            ("read", "Reader"),
                            (
                                "read-without-comments",
                                "Reader without comment access",
                            ),
                        ],
                        max_length=21,
                    ),
                ),
                (
                    "document",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="document.Document",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="DocumentRevision",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "note",
                    models.CharField(blank=True, default="", max_length=255),
                ),
                ("date", models.DateTimeField(auto_now=True)),
                (
                    "file_object",
                    models.FileField(
                        upload_to=document.models.revision_filename
                    ),
                ),
                (
                    "file_name",
                    models.CharField(blank=True, default="", max_length=255),
                ),
                (
                    "document",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="document.Document",
                    ),
                ),
                (
                    "doc_version",
                    models.DecimalField(
                        decimal_places=1, default=3.1, max_digits=3
                    ),
                ),
            ],
        ),
    ]



import os
import time
from tempfile import mkdtemp

from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC

from django.conf import settings

from testing.selenium_helper import SeleniumHelper
from channels.testing import ChannelsLiveServerTestCase


class BibliographyOverviewTest(SeleniumHelper, ChannelsLiveServerTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.download_dir = mkdtemp()
        driver_data = cls.get_drivers(1, cls.download_dir)
        cls.driver = driver_data["drivers"][0]
        cls.client = driver_data["clients"][0]
        cls.driver.implicitly_wait(driver_data["wait_time"])
        cls.wait_time = driver_data["wait_time"]

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()
        os.rmdir(cls.download_dir)
        super().tearDownClass()

    def setUp(self):
        self.base_url = self.live_server_url
        self.verificationErrors = []
        self.accept_next_alert = True
        self.user = self.create_user(
            username="Yeti", email="yeti@snowman.com", passtext="otter1"
        )
        self.login_user(self.user, self.driver, self.client)

    def test_overview(self):
        driver = self.driver
        driver.get(self.base_url + "/bibliography/")
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'All categories'])[1]/following::button[1]",
        ).click()
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[1]",
        ).click()
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[1]",
        ).clear()
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[1]",
        ).send_keys("Fish")
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::span[3]",
        ).click()
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[2]",
        ).click()
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[2]",
        ).clear()
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[2]",
        ).send_keys("Table")
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::span[4]",
        ).click()
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[3]",
        ).click()
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[3]",
        ).clear()
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::input[3]",
        ).send_keys("Jungle")
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Edit Categories'])[1]/following::button[2]",
        ).click()
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'All categories'])[1]/following::button[1]",
        ).click()
        try:
            self.assertEqual(
                "Fish",
                driver.find_element(By.ID, "categoryTitle_1").get_attribute(
                    "value"
                ),
            )
        except AssertionError as e:
            self.verificationErrors.append(str(e))
        try:
            self.assertEqual(
                "Table",
                driver.find_element(By.ID, "categoryTitle_2").get_attribute(
                    "value"
                ),
            )
        except AssertionError as e:
            self.verificationErrors.append(str(e))
        driver.find_element(
            By.CSS_SELECTOR,
            "button.fw-dark",
        ).click()
        driver.find_element(
            By.CSS_SELECTOR,
            "button[title='Register new source (Alt-n)']",
        ).click()
        driver.find_element(By.ID, "select-bibtype").click()
        driver.find_element(
            By.CSS_SELECTOR, "#select-bibtype option[value=article]"
        ).click()
        title_of_publication = driver.find_element(
            By.CSS_SELECTOR, ".journaltitle .ProseMirror"
        )
        title_of_publication.click()
        title_of_publication.send_keys("Title of publication")
        title = driver.find_element(By.CSS_SELECTOR, ".title .ProseMirror")
        title.click()
        title.send_keys("The title")
        first_name = driver.find_element(
            By.CSS_SELECTOR, ".given .ProseMirror"
        )
        first_name.click()
        first_name.click()  # TODO: We should not need two clicks
        first_name.send_keys("Hans")
        last_name = driver.find_element(
            By.CSS_SELECTOR, ".family .ProseMirror"
        )
        last_name.click()
        last_name.send_keys("Hansen")
        date_field = driver.find_element(By.CSS_SELECTOR, "input.date")
        date_field.click()
        date_field.send_keys("1984")
        driver.find_element(By.LINK_TEXT, "Categories").click()
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Categories'])[2]/following::div[1]",
        ).click()
        driver.find_element(
            By.XPATH,
            "(.//*[normalize-space(text()) and normalize-space(.)="
            "'Table'])[1]/following::div[1]",
        ).click()
        try:
            self.assertEqual(
                "Jungle",
                driver.find_element(
                    By.XPATH,
                    "(.//*[normalize-space(text()) and normalize-space(.)="
                    "'Table'])[1]/following::div[1]",
                ).text,
            )
        except AssertionError as e:
            self.verificationErrors.append(str(e))
        driver.find_element(By.CSS_SELECTOR, "button.fw-dark").click()
        # Make change to citation source
        driver.find_element(By.CSS_SELECTOR, ".edit-bib").click()
        date_input = driver.find_element(
            By.CSS_SELECTOR, ".entry-field.date input"
        )
        date_input.click()
        date_input.send_keys(Keys.BACKSPACE)
        date_input.send_keys("5")
        driver.find_element(By.CSS_SELECTOR, "button.fw-dark").click()
        # Closed citation dialog
        search_input = driver.find_element(
            By.CSS_SELECTOR,
            ".fw-overview-menu-item .fw-button input[type=search]",
        )
        search_input.click()
        search_input.send_keys("women")
        try:
            self.assertEqual(
                "No sources found",
                driver.find_element(
                    By.CSS_SELECTOR,
                    ".fw-data-table.fw-large.datatable-table td",
                ).text,
            )
        except AssertionError as e:
            self.verificationErrors.append(str(e))
        search_input.send_keys(Keys.BACKSPACE)
        search_input.send_keys(Keys.BACKSPACE)
        search_input.send_keys(Keys.BACKSPACE)
        search_input.send_keys(Keys.BACKSPACE)
        search_input.send_keys(Keys.BACKSPACE)
        try:
            self.assertEqual(
                "The title",
                driver.find_element(
                    By.CSS_SELECTOR, ".fw-data-table-title .edit-bib"
                ).text,
            )
        except AssertionError as e:
            self.verificationErrors.append(str(e))
        driver.find_element(
            By.CSS_SELECTOR, "button[title='Upload BibTeX file (Alt-u)']"
        ).click()
        # bibliography path
        bib_path = os.path.join(
            settings.PROJECT_PATH,
            "bibliography/tests/uploads/bibliography.bib",
        )
        driver.find_element(By.ID, "bib-uploader").send_keys(bib_path)
        driver.find_element(By.CSS_SELECTOR, "button.submit-import").click()
        book_title_el = WebDriverWait(driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, ".fw-data-table tr:nth-child(2) .edit-bib")
            )
        )
        self.assertEqual(
            "Lean UX: Applying lean principles to improve user experience",
            book_title_el.text,
        )

        # Export through dropdown menu
        self.driver.find_element(
            By.CSS_SELECTOR, "tr:nth-child(1) > td > label"
        ).click()
        driver.find_element(By.CSS_SELECTOR, ".dt-bulk-dropdown").click()
        driver.find_element(
            By.CSS_SELECTOR, "li.content-menu-item:nth-child(2)"
        ).click()
        time.sleep(1)
        assert os.path.isfile(
            os.path.join(self.download_dir, "bibliography.zip")
        )
        os.remove(os.path.join(self.download_dir, "bibliography.zip"))

        # Delete bib entry
        entries = self.driver.find_elements(
            By.CSS_SELECTOR, ".fw-data-table .edit-bib"
        )
        self.assertEqual(len(entries), 2)
        driver.find_element(By.CSS_SELECTOR, ".delete-bib").click()
        driver.find_element(By.CSS_SELECTOR, "button.fw-dark").click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.staleness_of(entries[0])
        )
        entries = self.driver.find_elements(
            By.CSS_SELECTOR, ".fw-data-table .edit-bib"
        )
        self.assertEqual(len(entries), 1)

        # Delete through dropdown menu
        self.driver.find_element(
            By.CSS_SELECTOR, "tr:nth-child(1) > td > label"
        ).click()
        driver.find_element(By.CSS_SELECTOR, ".dt-bulk-dropdown").click()
        driver.find_element(
            By.CSS_SELECTOR, "li.content-menu-item:nth-child(1)"
        ).click()
        driver.find_element(By.CSS_SELECTOR, "button.fw-dark").click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.staleness_of(entries[0])
        )
        entries = self.driver.find_elements(
            By.CSS_SELECTOR, ".fw-data-table .edit-bib"
        )
        self.assertEqual(len(entries), 0)
        # Delete category
        driver.find_element(
            By.CSS_SELECTOR, "button[title='Edit categories (Alt-e)']"
        ).click()
        category_inputs = self.driver.find_elements(
            By.CSS_SELECTOR, "#editCategoryList tr"
        )
        self.assertEqual(len(category_inputs), 4)
        driver.find_element(
            By.CSS_SELECTOR, ".fw-add-input.icon-addremove"
        ).click()
        driver.find_element(By.CSS_SELECTOR, "button.fw-dark").click()
        driver.find_element(
            By.CSS_SELECTOR, "button[title='Edit categories (Alt-e)']"
        ).click()
        category_inputs = self.driver.find_elements(
            By.CSS_SELECTOR, "#editCategoryList tr"
        )
        self.assertEqual(len(category_inputs), 3)

    def tearDown(self):
        self.assertEqual([], self.verificationErrors)
        return super().tearDown()



import json
from django.db import migrations, models


def text_to_json(apps, schema_editor):
    Entry = apps.get_model("bibliography", "Entry")
    entries = Entry.objects.all()
    for entry in entries:
        entry.cats = json.loads(entry.entry_cat)
        entry.fields = json.loads(entry.fields_text)
        entry.save()


def json_to_text(apps, schema_editor):
    Entry = apps.get_model("bibliography", "Entry")
    entries = Entry.objects.all()
    for entry in entries:
        entry.entry_cat = json.dumps(entry.cats)
        entry.fields_text = json.dumps(entry.fields)
        entry.save()


class Migration(migrations.Migration):
    dependencies = [
        ("bibliography", "0001_squashed_0011_auto_20170101_1647"),
    ]

    operations = [
        migrations.AddField(
            model_name="entry",
            name="cats",
            field=models.JSONField(default=list),
        ),
        migrations.RenameField(
            model_name="entry", old_name="fields", new_name="fields_text"
        ),
        migrations.AddField(
            model_name="entry",
            name="fields",
            field=models.JSONField(default=dict),
        ),
        migrations.RunPython(text_to_json, json_to_text),
        migrations.RemoveField(
            model_name="entry",
            name="entry_cat",
        ),
        migrations.RemoveField(
            model_name="entry",
            name="fields_text",
        ),
    ]



from time import mktime
import json

from django.contrib.auth.decorators import login_required
from django.http import JsonResponse
from django.core.serializers.python import Serializer
from django.utils.translation import gettext as _
from django.views.decorators.http import require_POST
from django.conf import settings

from base.decorators import ajax_required
from usermedia.models import Image, ImageCategory, UserImage
from .models import ALLOWED_FILETYPES


class SimpleSerializer(Serializer):
    def end_object(self, obj):
        self._current["id"] = obj._get_pk_val()
        self.objects.append(self._current)


serializer = SimpleSerializer()


# save changes or create a new entry
@login_required
@ajax_required
@require_POST
def save(request):
    response = {}
    response["errormsg"] = {}
    if (
        "image" in request.FILES
        and request.FILES["image"].content_type not in ALLOWED_FILETYPES
    ):
        status = 200  # Not implemented
        response["errormsg"]["error"] = _("Filetype not supported")
    elif (
        settings.MEDIA_MAX_SIZE
        and "image" in request.FILES
        and request.FILES["image"].size > settings.MEDIA_MAX_SIZE
    ):
        status = 413
        response["errormsg"]["error"] = _("Image too large")
    else:
        image = False
        if "id" in request.POST and "image" not in request.FILES:
            user_image = UserImage.objects.filter(
                image_id=int(request.POST["id"]), owner=request.user
            ).first()
            if user_image:
                image = user_image.image
                status = 200
        if image is False:
            image = Image()
            image.uploader = request.user
            user_image = UserImage()
            user_image.owner = request.user
            status = 201
            if "checksum" in request.POST:
                image.checksum = request.POST["checksum"]
        user_image.title = request.POST["title"]
        if "copyright" in request.POST:
            user_image.copyright = json.loads(request.POST["copyright"])
        if "cats" in request.POST:
            user_image.cats = json.loads(request.POST["cats"])
        if "image" in request.FILES:
            image.image = request.FILES["image"]
        if status == 201 and "image" not in request.FILES:
            status = 200
            response["errormsg"]["error"] = _("No file uploaded")
        else:
            image.save()
            user_image.image = image
            user_image.save()
            response["values"] = {
                "id": image.id,
                "title": user_image.title,
                "copyright": user_image.copyright,
                "image": image.image.url,
                "file_type": image.file_type,
                "added": mktime(image.added.timetuple()) * 1000,
                "checksum": image.checksum,
                "cats": user_image.cats,
            }
            if image.thumbnail:
                response["values"]["thumbnail"] = image.thumbnail.url
                response["values"]["height"] = image.height
                response["values"]["width"] = image.width
    return JsonResponse(response, status=status)


# delete an image
@login_required
@ajax_required
@require_POST
def delete(request):
    response = {}
    status = 201
    ids = request.POST.getlist("ids[]")
    UserImage.objects.filter(image_id__in=ids, owner=request.user).delete()
    for image in Image.objects.filter(id__in=ids):
        if image.is_deletable():
            image.delete()
    return JsonResponse(response, status=status)


# returns list of images
@login_required
@ajax_required
@require_POST
def images(request):
    response = {}
    status = 200
    response["imageCategories"] = serializer.serialize(
        ImageCategory.objects.filter(category_owner=request.user)
    )
    response["images"] = []
    user_images = UserImage.objects.filter(owner=request.user).select_related(
        "image"
    )
    for user_image in user_images:
        image = user_image.image
        if image.image:
            field_obj = {
                "id": image.id,
                "title": user_image.title,
                "copyright": user_image.copyright,
                "image": image.image.url,
                "file_type": image.file_type,
                "added": mktime(image.added.timetuple()) * 1000,
                "checksum": image.checksum,
                "cats": user_image.cats,
            }
            if image.thumbnail:
                field_obj["thumbnail"] = image.thumbnail.url
                field_obj["height"] = image.height
                field_obj["width"] = image.width
            response["images"].append(field_obj)
    return JsonResponse(response, status=status)


# save changes or create a new category
@login_required
@ajax_required
@require_POST
def save_category(request):
    response = {}
    response["entries"] = []
    ids = request.POST.getlist("ids[]")
    titles = request.POST.getlist("titles[]")
    ImageCategory.objects.filter(category_owner=request.user).exclude(
        id__in=ids
    ).delete()
    x = 0
    for the_id in ids:
        the_id = int(the_id)
        the_title = titles[x]
        x += 1
        if 0 == the_id:
            # if the category is new, then create new
            the_cat = ImageCategory(
                category_title=the_title, category_owner=request.user
            )
        else:
            # if the category already exists, update the title
            the_cat = ImageCategory.objects.get(pk=the_id)
            the_cat.category_title = the_title
        the_cat.save()
        response["entries"].append(
            {"id": the_cat.id, "category_title": the_cat.category_title}
        )
    status = 201

    return JsonResponse(response, status=status)



# Generated by Django 4.1.10 on 2023-09-28 09:57

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0010_alter_user_contacts"),
    ]

    operations = [
        migrations.DeleteModel(
            name="LoginAs",
        ),
        migrations.AlterModelOptions(
            name="user",
            options={
                "permissions": (("can_login_as", "Can login as another user"),)
            },
        ),
    ]



from django.db import models
from django.utils.translation import gettext as _


class DocumentStyle(models.Model):
    title = models.CharField(
        max_length=128,
        help_text="The human readable title.",
        default="Default",
    )
    slug = models.SlugField(
        max_length=20,
        help_text="The base of the filenames the style occupies.",
        default="default",
    )
    contents = models.TextField(
        help_text="The CSS style definiton.", default=""
    )
    document_template = models.ForeignKey(
        "document.DocumentTemplate", on_delete=models.deletion.CASCADE
    )

    def __str__(self):
        return self.title

    class Meta(object):
        unique_together = (("slug", "document_template"),)


def documentstylefile_location(instance, filename):
    # preserve the original filename
    instance.filename = filename
    return "/".join(["style-files", filename])


class DocumentStyleFile(models.Model):
    file = models.FileField(
        upload_to=documentstylefile_location,
        help_text=(
            "A file references in the style. The filename will be replaced "
            "with the final url of the file in the style."
        ),
    )
    filename = models.CharField(
        max_length=255, help_text="The original filename."
    )
    style = models.ForeignKey(
        "DocumentStyle", on_delete=models.deletion.CASCADE
    )

    def __str__(self):
        return self.filename + " of " + self.style.title

    def natural_key(self):
        return (self.file.url, self.filename)

    class Meta(object):
        unique_together = (("filename", "style"),)


TEMPLATE_CHOICES = (("docx", "DOCX"), ("odt", "ODT"))


def template_filename(instance, filename):
    instance.title = filename.split(".")[0]
    return "/".join(["export-template-files", filename])


class ExportTemplate(models.Model):
    template_file = models.FileField(upload_to=template_filename)
    title = models.CharField(
        max_length=128,
        help_text="The human readable title.",
        default=_("Default"),
    )
    file_type = models.CharField(
        max_length=5, choices=TEMPLATE_CHOICES, blank=False
    )
    document_template = models.ForeignKey(
        "document.DocumentTemplate", on_delete=models.deletion.CASCADE
    )

    def __str__(self):
        return self.title + " (" + self.file_type + ")"

    class Meta(object):
        unique_together = (("title", "document_template"),)



from django.core.mail import send_mail
from django.conf import settings
from django.utils.translation import gettext as _

from base.html_email import html_email


def send_share_notification(
    document_title,
    owner,
    link,
    collaborator_name,
    collaborator_email,
    rights,
    change,
):
    if len(document_title) == 0:
        document_title = _("Untitled")
    if change:
        message_text = _(
            (
                "Hey %(collaborator_name)s,\n%(owner)s has changed your access "
                "rights to %(rights)s on the document '%(document_title)s'. "
                "\nOpen the document: %(link)s"
            )
        ) % {
            "owner": owner,
            "rights": rights,
            "collaborator_name": collaborator_name,
            "link": link,
            "document_title": document_title,
        }
        body_html_intro = _(
            (
                "<p>Hey %(collaborator_name)s,<br>%(owner)s has changed your "
                "access rights to %(rights)s on the document "
                "'%(document_title)s'.</p>"
            )
        ) % {
            "owner": owner,
            "rights": rights,
            "collaborator_name": collaborator_name,
            "document_title": document_title,
        }
    else:
        message_text = _(
            (
                "Hey %(collaborator_name)s,\n%(owner)s has shared the document "
                "'%(document_title)s' with you and given you %(rights)s access "
                "rights. "
                "\nOpen document: %(link)s"
            )
        ) % {
            "owner": owner,
            "rights": rights,
            "collaborator_name": collaborator_name,
            "link": link,
            "document_title": document_title,
        }
        body_html_intro = _(
            (
                "<p>Hey %(collaborator_name)s,<br>%(owner)s has shared the "
                "document '%(document_title)s' with you and given you "
                "%(rights)s access rights.</p>"
            )
        ) % {
            "owner": owner,
            "rights": rights,
            "collaborator_name": collaborator_name,
            "document_title": document_title,
        }

    body_html = (
        "<h1>%(document_title)s %(shared)s</h1>"
        "%(body_html_intro)s"
        "<table>"
        "<tr><td>"
        "%(Document)s"
        "</td><td>"
        "<b>%(document_title)s</b>"
        "</td></tr>"
        "<tr><td>"
        "%(Author)s"
        "</td><td>"
        "%(owner)s"
        "</td></tr>"
        "<tr><td>"
        "%(AccessRights)s"
        "</td><td>"
        "%(rights)s"
        "</td></tr>"
        "</table>"
        '<div class="actions"><a class="button" href="%(link)s">'
        "%(AccessTheDocument)s"
        "</a></div>"
    ) % {
        "shared": _("shared"),
        "body_html_intro": body_html_intro,
        "Document": _("Document"),
        "document_title": document_title,
        "Author": _("Author"),
        "owner": owner,
        "AccessRights": _("Access Rights"),
        "rights": rights,
        "link": link,
        "AccessTheDocument": _("Access the document"),
    }
    send_mail(
        _("Document shared: %(document_title)s")
        % {"document_title": document_title},
        message_text,
        settings.DEFAULT_FROM_EMAIL,
        [collaborator_email],
        fail_silently=True,
        html_message=html_email(body_html),
    )


def send_comment_notification(
    notification_type,
    commentator,
    collaborator_name,
    collaborator_email,
    link,
    document_title,
    comment_text,
    comment_html,
):
    if notification_type == "mention":
        message_text = _(
            (
                "Hey %(collaborator_name)s,\n%(commentator)s has mentioned you "
                "in a comment in the document '%(document)s':"
                "\n\n%(comment_text)s"
                "\n\nGo to the document here: %(link)s"
            )
        ) % {
            "commentator": commentator,
            "collaborator_name": collaborator_name,
            "link": link,
            "document": document_title,
            "comment_text": comment_text,
        }

        body_html_title = _(
            (
                "Hey %(collaborator_name)s,<br>%(commentator)s has mentioned "
                "you in a comment in the document '%(document_title)s'."
            )
        ) % {
            "commentator": commentator,
            "collaborator_name": collaborator_name,
            "document_title": document_title,
        }
        message_title = _("Comment on : %(document_title)s") % {
            "document_title": document_title
        }
    else:
        message_text = _(
            (
                "Hey %(collaborator_name)s,\n%(commentator)s has assigned you to "
                "a comment in the document '%(document)s':\n\n%(comment_text)s"
                "\n\nGo to the document here: %(link)s"
            )
        ) % {
            "commentator": commentator,
            "collaborator_name": collaborator_name,
            "link": link,
            "document": document_title,
            "comment_text": comment_text,
        }
        body_html_title = _(
            (
                "Hey %(collaborator_name)s,<br>%(commentator)s has assigned you "
                "to a comment in the document '%(document_title)s'."
            )
        ) % {
            "commentator": commentator,
            "collaborator_name": collaborator_name,
            "document_title": document_title,
        }
        message_title = _("Comment assignment on : %(document_title)s") % {
            "document_title": document_title
        }

    body_html = _(
        (
            "<p>Hey %(collaborator_name)s,<br>%(commentator)s has assigned "
            "you to a comment in the document '%(document)s':</p>"
            "%(comment_html)s"
            '<p>Go to the document <a href="%(link)s">here</a>.</p>'
        )
    ) % {
        "commentator": commentator,
        "collaborator_name": collaborator_name,
        "link": link,
        "document": document_title,
        "comment_html": comment_html,
    }

    body_html = (
        "<h1>%(body_html_title)s</h1>"
        "<table>"
        "<tr><td>"
        "%(Document)s"
        "</td><td>"
        "<b>%(document_title)s</b>"
        "</td></tr>"
        "<tr><td>"
        "%(Author)s"
        "</td><td>"
        "%(commentator)s"
        "</td></tr>"
        "<tr><td>"
        "%(Comment)s"
        "</td><td>"
        "%(comment_html)s"
        "</td></tr>"
        "</table>"
        '<div class="actions"><a class="button" href="%(link)s">'
        "%(AccessTheDocument)s"
        "</a></div>"
    ) % {
        "body_html_title": body_html_title,
        "Document": _("Document"),
        "document_title": document_title,
        "Author": _("Author"),
        "commentator": commentator,
        "Comment": _("Comment"),
        "comment_html": comment_html,
        "link": link,
        "AccessTheDocument": _("Access the document"),
    }
    send_mail(
        message_title,
        message_text,
        settings.DEFAULT_FROM_EMAIL,
        [collaborator_email],
        fail_silently=True,
        html_message=html_email(body_html),
    )



from django.conf import settings
from django.core.mail import send_mail


def send_feedback(from_sender, reply_to, message):
    send_mail(
        f"Feedback from {from_sender} {reply_to}",
        message,
        settings.DEFAULT_FROM_EMAIL,
        [settings.CONTACT_EMAIL],
        fail_silently=True,
    )



import os
import shelve
import time

from django.core.mail.backends.base import BaseEmailBackend
from django.conf import settings

import logging

logger = logging.getLogger(__name__)

# Source: https://github.com/jricardo27/channels_example/commit/637b689870c2f879286f30eeb0c3e1d65283d557#diff-180eb8ba4002fc012b871a9b0c0ba23b31727c34a20d8470421841e9e3bb637b

MAIL_STORAGE_BASE = ".test_mail_storage_"
OUTBOX = "outbox"


class EmailBackend(BaseEmailBackend):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if hasattr(settings, "MAIL_STORAGE_NAME"):
            self.mail_storage = MAIL_STORAGE_BASE + settings.MAIL_STORAGE_NAME
        else:
            self.mail_storage = MAIL_STORAGE_BASE

    def send_messages(self, messages):
        """Redirect messages to the dummy outbox"""
        logger.debug(f"send_messages: {len(messages)}")
        storage = shelve.open(self.mail_storage)
        outbox = storage.get(OUTBOX, [])
        storage.close()
        msg_count = 0
        for message in messages:  # .message() triggers header validation
            message.message()
            outbox.append(message)
            msg_count += 1
        storage = shelve.open(self.mail_storage)
        storage[OUTBOX] = outbox
        storage.close()
        logger.debug(f"Total length: {len(outbox)}")
        return msg_count


def get_outbox(mail_storage_name=None):
    time.sleep(1)
    if mail_storage_name:
        mail_storage = MAIL_STORAGE_BASE + mail_storage_name
    else:
        mail_storage = MAIL_STORAGE_BASE
    storage = shelve.open(mail_storage)
    outbox = storage.get(OUTBOX, [])
    storage.close()
    logger.info(f"Mailbox length: {len(outbox)}")
    return outbox


def empty_outbox(mail_storage_name=None):
    if mail_storage_name:
        mail_storage = MAIL_STORAGE_BASE + mail_storage_name
    else:
        mail_storage = MAIL_STORAGE_BASE
    storage = shelve.open(mail_storage)
    storage[OUTBOX] = []
    storage.close()


def delete_outbox(mail_storage_name=None):
    if mail_storage_name:
        mail_storage = MAIL_STORAGE_BASE + mail_storage_name
    else:
        mail_storage = MAIL_STORAGE_BASE
    # Clean outbox.
    try:
        os.remove(mail_storage)
    except OSError:
        pass



import ldap3

from allauth.socialaccount.models import SocialAccount
from django.conf import settings

from base.management import BaseCommand


def init_ldap():
    server = ldap3.Server("ldaps://ads.mwn.de")
    connection = ldap3.Connection(
        server,
        f"CN={settings.LDAP_USER},OU=Users,ou=TU,ou=IAM,dc=ads,dc=mwn,dc=de",
        settings.LDAP_PASSWORD,
        auto_bind=True,
    )
    return connection


def check_user_in_ldap(connection, uid):
    return connection.search(
        "ou=Users,ou=TU,ou=IAM,dc=ads,dc=mwn,dc=de", f"(uid={uid})"
    )


class Command(BaseCommand):
    help = "Verify state of users in Active Directory. Deactivate old users."

    def add_arguments(self, parser):
        parser.add_argument(
            "--delete",
            action="store_true",
            dest="delete",
            default=False,
            help="Whether to delete instead of deactivate old users.",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            dest="dryrun",
            default=False,
            help="Whether to check users in LDAP directory without actually changing them.",
        )
        parser.add_argument(
            "--verbose",
            action="store_true",
            dest="verbose",
            default=False,
            help="Whether to print actions to command line.",
        )

    def handle(self, *args, **options):
        if not (
            hasattr(settings, "LDAP_USER")
            and hasattr(settings, "LDAP_PASSWORD")
        ):
            self.stdout.write(
                "Please set LDAP_USER and LDAP_PASSWORD in configuration.py."
            )
            return
        connection = init_ldap()
        if not connection.bound:
            self.stdout.write(
                "Connection to LDAP server could not be established."
            )
            return
        activate_count = 0
        delete_count = 0
        deactivate_count = 0

        for sa in SocialAccount.objects.all():
            if check_user_in_ldap(connection, sa.uid):
                if not sa.user.is_active:
                    if options["verbose"]:
                        self.stdout.write(f"Activating {sa.user.username}.")
                    if not options["dryrun"]:
                        sa.user.is_active = True
                        sa.user.save()
                    activate_count += 1
            elif options["delete"]:
                if options["verbose"]:
                    self.stdout.write(f"Deleting {sa.user.username}.")
                if not options["dryrun"]:
                    sa.user.delete()
                delete_count += 1
            elif sa.user.is_active:
                if options["verbose"]:
                    self.stdout.write(f"Deactivating {sa.user.username}.")
                if not options["dryrun"]:
                    sa.user.is_active = False
                    sa.user.save()
                deactivate_count += 1
        connection.unbind()
        if not options["dryrun"]:
            self.stdout.write(
                f"Activated: {activate_count}, Deactivated: {deactivate_count}, Deleted: {delete_count}, Verified: {len(SocialAccount.objects.all())}"
            )



from typing import Tuple

from document.models import AccessRight, Document, DocumentTemplate


class SessionUserInfo:
    """
    Class for storing information about users in session
    author: akorovin
    """

    def __init__(self, current_user):
        self.user = current_user
        self.is_owner = False
        self.access_rights = "read"
        self.document_id = 0
        self.path = ""
        self.path_object = None

    async def create_doc(self, template_id):
        """
        Asynchronously create a new document
        """
        # Get the template asynchronously
        template = await DocumentTemplate.objects.filter(
            id=int(template_id)
        ).afirst()

        # Create document asynchronously
        if template:
            document = await Document.objects.acreate(
                owner_id=self.user.id, template=template
            )
        else:
            document = await Document.objects.acreate(owner_id=self.user.id)

        return document

    async def init_access(self, document_id) -> Tuple[Document, bool]:
        """
        Asynchronously initializes access to document by id

        :param document_id: The ID of the document to access
        :return: Returns document and bool value that user can access
        :rtype: tuple
        """
        can_access = False

        # Use select_related to prefetch owner and template in one query
        document = (
            await Document.objects.select_related("owner", "template")
            .filter(id=int(document_id))
            .afirst()
        )

        if document is None:
            return (False, False)
        else:
            self.document_id = document.id

            # Now document.owner is already loaded, so this won't trigger a new query
            if document.owner.id == self.user.id:
                self.access_rights = "write"
                self.is_owner = True
                self.path = document.path
                self.path_object = document
                can_access = True
            else:
                self.is_owner = False

                # Get access rights asynchronously with prefetched relations
                access_right = await AccessRight.objects.filter(
                    document_id=document.id, user=self.user
                ).afirst()

                if access_right:
                    self.access_rights = access_right.rights
                    self.path = access_right.path
                    self.path_object = access_right
                    can_access = True

        return (document, can_access)



import json
import random
from httpx_ws import connect_ws
from django.contrib.auth import get_user_model
from django.conf import settings
from django.shortcuts import render
from django.views.decorators.csrf import ensure_csrf_cookie
from django.contrib.sites.shortcuts import get_current_site
from django.contrib.flatpages.models import FlatPage
from django.http import JsonResponse
from django.views.decorators.http import require_GET, require_POST
from django.contrib.admin.views.decorators import staff_member_required

from allauth.socialaccount.adapter import get_adapter

from user.helpers import Avatars
from .decorators import ajax_required
from . import get_version
from .helpers.ws import get_url_base
from .models import Presence


@ensure_csrf_cookie
def app(request):
    """
    Load a page controlled by the JavaScript app.
    Used all user facing pages after login.
    """
    return render(request, "app.html", {"version": get_version()})


def api_404(request):
    """
    Show a 404 error within the API.
    """
    return render(request, "api_404.html", status=404)


@ajax_required
@require_POST
def configuration(request):
    """
    Load the configuration options of the page that are request dependent.
    """
    if len(settings.PORTS) < 2:
        ws_url_base = "/ws"
    else:
        # Get the Origin header safely with a fallback
        origin = request.headers.get("Origin")
        if not origin:
            # Use the current host as fallback
            protocol = "https" if request.is_secure() else "http"
            origin = f"{protocol}://{request.get_host()}"
        ws_url_base = get_url_base(origin, random.choice(settings.PORTS))
    socialaccount_providers = []
    for provider in get_adapter(request).list_providers(request):
        socialaccount_providers.append(
            {
                "id": provider.id,
                "name": provider.name,
                "login_url": provider.get_login_url(request),
            }
        )
    response = {
        "language": request.LANGUAGE_CODE,
        "socialaccount_providers": socialaccount_providers,
        "ws_url_base": ws_url_base,
    }
    user = request.user
    if user.is_authenticated:
        user = (
            get_user_model()
            .objects.prefetch_related("emailaddress_set", "socialaccount_set")
            .get(id=user.id)
        )
        avatars = Avatars()
        response["user"] = {
            "id": user.id,
            "username": user.username,
            "first_name": user.first_name,
            "name": user.readable_name,
            "last_name": user.last_name,
            "language": user.language,
            "avatar": avatars.get_url(user),
            "emails": [],
            "socialaccounts": [],
            "is_authenticated": True,
        }

        for emailaddress in user.emailaddress_set.all():
            email = {
                "address": emailaddress.email,
            }
            if emailaddress.primary:
                email["primary"] = True
            if emailaddress.verified:
                email["verified"] = True
            response["user"]["emails"].append(email)
        for account in user.socialaccount_set.all():
            try:
                provider_account = account.get_provider_account()
                response["user"]["socialaccounts"].append(
                    {
                        "id": account.id,
                        "provider": account.provider,
                        "name": provider_account.to_str(),
                    }
                )
            except KeyError:
                # Social account provider has been removed.
                pass
        response["user"]["waiting_invites"] = user.invites_to.exists()

    else:
        response["user"] = {"is_authenticated": False}
    return JsonResponse(response, status=200)


def manifest_json(request):
    """
    Load the manifest.json.
    """
    return render(request, "manifest.json")


# view is shown only in admin interface, so authentication is taken care of
def admin_console(request):
    """
    Load the admin console page.
    """
    return render(request, "admin/console.html")


@ajax_required
@require_GET
@staff_member_required
def connection_info(request):
    """
    Return info about currently connected clients.
    """
    response = {}
    Presence.prune()
    response["sessions"] = Presence.objects.count()
    response["users"] = (
        Presence.objects.values_list("user", flat=True).distinct().count()
    )
    return JsonResponse(response, status=200)


def send_to_server(server_url, message, headers):
    with connect_ws(
        server_url,
        headers={
            "Origin": headers["Origin"],
            "Cookie": headers["Cookie"],
            "User-Agent": "Fidus Writer",
        },
    ) as websocket:
        websocket.send_text(
            json.dumps(
                {"type": "system_message", "message": message, "s": 1, "c": 1}
            )
        )


@ajax_required
@require_POST
@staff_member_required
def send_system_message(request):
    """
    Send out a system message to all clients connected to the frontend.
    """
    response = {}
    message = request.POST["message"]

    servers = set(
        Presence.objects.values_list("server_url", flat=True).distinct()
    )

    for server in servers:
        send_to_server(server, message, dict(request.headers))

    return JsonResponse(response, status=200)


@ajax_required
@require_POST
def flatpage(request):
    """
    Models: `flatpages.flatpages`
    Context:
        flatpage
            `flatpages.flatpages` object
    """
    response = {}
    status = 404
    url = request.POST["url"]
    site_id = get_current_site(request).id
    flatpage = FlatPage.objects.filter(url=url, sites=site_id).first()
    if flatpage:
        status = 200
        response["title"] = flatpage.title
        response["content"] = flatpage.content
    return JsonResponse(response, status=status)



# Generated by Django 3.1.4 on 2021-05-27 10:18

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("document", "0015_migrate_invites"),
    ]

    operations = [
        migrations.DeleteModel(
            name="AccessRightInvite",
        ),
    ]



# Generated by Django 3.1.2 on 2020-11-09 01:55

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("document", "0004_move_json_data"),
    ]

    operations = [
        migrations.AlterField(
            model_name="document",
            name="bibliography",
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name="document",
            name="comments",
            field=models.JSONField(blank=True, default=dict),
        ),
    ]



from django.core import serializers
from django.db import models
from asgiref.sync import sync_to_async

PythonSerializer = serializers.get_serializer("python")


class PythonWithURLSerializer(PythonSerializer):
    def handle_field(self, obj, field):
        value = field.value_from_object(obj)
        if isinstance(field, models.FileField) and hasattr(value, "url"):
            self._current[field.name] = value.url
        else:
            return super().handle_field(obj, field)

    def end_object(self, obj):
        # We add reverse foreign key relations if they are explicitly added.
        # Needed to include DocumentStyleFont with DocumentStyle
        if self.selected_fields is not None:
            for field in obj._meta.related_objects:
                if not field.one_to_many:
                    continue
                if field.related_name:
                    related_name = field.related_name
                else:
                    related_name = f"{field.name}_set"
                if related_name not in self.selected_fields:
                    continue
                if self.use_natural_foreign_keys and hasattr(
                    field.remote_field.model, "natural_key"
                ):

                    def o2m_value(value):
                        return value.natural_key()

                else:

                    def o2m_value(value):
                        return self._value_from_field(value, value._meta.pk)

                self._current[related_name] = [
                    o2m_value(related)
                    for related in getattr(obj, related_name).iterator()
                ]

        return super().end_object(obj)

    # Add an async version for Django 5.1 compatibility
    async def aserialize(self, queryset, **options):
        """
        Asynchronous version of serialize method for use with async code
        """
        return await sync_to_async(self.serialize)(queryset, **options)



from django.contrib import admin
from django.shortcuts import render
from django.urls import path
from django.utils.translation import gettext as _
from . import models


class DocumentAdmin(admin.ModelAdmin):
    def get_urls(self):
        urls = super().get_urls()
        extra_urls = [
            path(
                "maintenance/",
                self.admin_site.admin_view(self.maintenance_view),
            )
        ]
        urls = extra_urls + urls
        return urls

    def maintenance_view(self, request):
        response = {}
        return render(request, "admin/document/maintenance.html", response)


admin.site.register(models.Document, DocumentAdmin)


class DocumentTemplateAdmin(admin.ModelAdmin):
    actions = [
        "duplicate",
    ]
    list_display = (
        "title",
        "user",
    )

    def duplicate(self, request, queryset):
        for template in queryset:
            document_styles = list(template.documentstyle_set.all())
            export_templates = list(template.exporttemplate_set.all())
            template.pk = None
            template.save()
            for ds in document_styles:
                style_files = list(ds.documentstylefile_set.all())
                ds.pk = None
                ds.document_template = template
                ds.save()
                for sf in style_files:
                    sf.pk = None
                    sf.style = ds
                    sf.save()
            for et in export_templates:
                et.pk = None
                et.document_template = template
                et.save()

    duplicate.short_description = _("Duplicate selected document templates")


admin.site.register(models.DocumentTemplate, DocumentTemplateAdmin)


class AccessRightAdmin(admin.ModelAdmin):
    pass


admin.site.register(models.AccessRight, AccessRightAdmin)


class DocumentRevisionAdmin(admin.ModelAdmin):
    pass


admin.site.register(models.DocumentRevision, DocumentRevisionAdmin)



from glob import glob
from setuptools import find_namespace_packages, setup
from setuptools.command.sdist import sdist as _sdist
from setuptools.command.bdist_egg import bdist_egg as _bdist_egg
from setuptools.command.install import install
from setuptools.command.build_py import build_py as _build_py
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
from babel.messages.frontend import compile_catalog as _compile_catalog


class compile_catalog(_compile_catalog):
    def initialize_options(self):
        super().initialize_options()
        self.domain = ["django", "djangojs"]
        self.directory = "fiduswriter/locale"


# From https://github.com/pypa/setuptools/pull/1574
class build_py(_build_py):
    def find_package_modules(self, package, package_dir):
        modules = super().find_package_modules(package, package_dir)
        patterns = self._get_platform_patterns(
            self.exclude_package_data,
            package,
            package_dir,
        )

        excluded_module_files = []
        for pattern in patterns:
            excluded_module_files.extend(glob(pattern))

        for f in excluded_module_files:
            for module in modules:
                if module[2] == f:
                    modules.remove(module)
        return modules


class bdist_egg(_bdist_egg):
    def run(self):
        self.run_command("compile_catalog")
        _bdist_egg.run(self)


class sdist(_sdist):
    """Custom build command."""

    def run(self):
        self.run_command("compile_catalog")
        _sdist.run(self)


class bdist_wheel(_bdist_wheel):
    """Custom build command."""

    def run(self):
        self.run_command("compile_catalog")
        _bdist_wheel.run(self)


cmdclass = {
    "compile_catalog": compile_catalog,
    "sdist": sdist,
    "build_py": build_py,
    "bdist_egg": bdist_egg,
    "bdist_wheel": bdist_wheel,
    "install": install,
}


setup(
    cmdclass=cmdclass,
    packages=find_namespace_packages(include=["fiduswriter"]),
    include_package_data=True,
    exclude_package_data={
        "": [
            "*.pyc",
            "*.pyo",
            "*.po",
            "*.pot",
            "*.db",
            ".*.db",
            "*.sqlite3",
            "ci/*",
            "build/*",
            "fiduswriter/media/*",
            "fiduswriter/.transpile/*",
            "fiduswriter/static-transpile/*",
            "fiduswriter/static-collected/*",
            "fiduswriter/static-libs/*",
            "fiduswriter/venv/*",
            "fiduswriter/media/*",
            "fiduswriter/static-transpile/*",
            "fiduswriter/static-libs/*",
            "fiduswriter/book/*",
            "fiduswriter/citation-api-import/*",
            "fiduswriter/languagetool/*",
            "fiduswriter/npm_mjs/*",
            "fiduswriter/ojs/*",
            "fiduswriter/phplist/*",
            "fiduswriter/payment/*",
            "fiduswriter/website/*",
            "fiduswriter/pandoc/*",
        ]
    },
)



# Generated by Django 4.1 on 2022-09-29 08:05

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0009_alter_user_contacts"),
    ]

    operations = [
        migrations.AlterField(
            model_name="user",
            name="contacts",
            field=models.ManyToManyField(
                blank=True, to=settings.AUTH_USER_MODEL
            ),
        ),
    ]



from urllib.parse import urlparse


def get_url_base(origin, conn):

    # Determine the expected values based on conn

    if isinstance(conn, int) and conn > 0:
        # conn is a port: expect origin's host and conn's port
        # Parse the origin URL (hostname and port)
        parsed_origin = urlparse(origin)
        origin_name = parsed_origin.hostname
        return f"{origin_name}:{conn}/ws"

    elif isinstance(conn, dict) and "external" in conn:
        # conn is a dict of type:
        # {"internal": 8000, "external": 9000}
        # or
        # {"internal": 8000, "external": "ws1.fiduswriter.com"}
        if isinstance(conn["external"], str):
            return f"{conn["external"]}/ws"
        elif isinstance(conn["external"], int):
            parsed_origin = urlparse(origin)
            origin_name = parsed_origin.hostname
            return f"{origin_name}:{conn["external"]}/ws"

    return "/ws"



#!/usr/bin/env python3
import os
import sys
from importlib import import_module


if "COVERAGE_PROCESS_START" in os.environ:
    import coverage

    coverage.process_startup()

SRC_PATH = os.path.dirname(os.path.realpath(__file__))
os.environ.setdefault("SRC_PATH", SRC_PATH)


def load_settings(settings_module):
    """
    Load settings from three levels, each overriding the previous one:
    1. Django's global_settings.py
    2. Fiduswriter's base settings.py
    3. User's configuration.py (or custom settings module)
    """
    from django.conf import global_settings as GLOBAL_SETTINGS
    from base import settings as BASE_SETTINGS

    settings_paths = [BASE_SETTINGS.__file__]

    # Start with Django's global settings
    settings_dict = {}
    for setting in dir(GLOBAL_SETTINGS):
        if setting.isupper():
            settings_dict[setting] = getattr(GLOBAL_SETTINGS, setting)

    # Override with Fiduswriter's base settings
    for setting in dir(BASE_SETTINGS):
        if setting.isupper():
            settings_dict[setting] = getattr(BASE_SETTINGS, setting)

    # Override with user's configuration settings if available
    try:
        config_module = import_module(settings_module)
        settings_paths.append(config_module.__file__)
        for setting in dir(config_module):
            if setting.isupper():
                settings_dict[setting] = getattr(config_module, setting)
    except ModuleNotFoundError:
        pass

    # Add metadata about settings
    settings_dict["SETTINGS_PATHS"] = settings_paths
    settings_dict["SETTINGS_MODULE"] = settings_module

    return settings_dict


def inner(default_project_path):
    """
    Main entry point for the Django management commands.

    Args:
        default_project_path: The default path to use if no --pythonpath is provided
    """
    sys.path.append(SRC_PATH)
    sys_argv = sys.argv
    PROJECT_PATH = False

    # Handle --pythonpath argument
    while "--pythonpath" in sys_argv:
        index = sys_argv.index("--pythonpath")
        PROJECT_PATH = os.path.join(os.getcwd(), sys_argv[index + 1])
        sys.path.insert(0, PROJECT_PATH)
        # We prevent the pythonpath to be handled later on by removing it from
        # sys_argv
        sys_argv = sys_argv[:index] + sys_argv[index + 2 :]

    if not PROJECT_PATH:
        PROJECT_PATH = default_project_path
        sys.path.insert(0, PROJECT_PATH)
    os.environ.setdefault("PROJECT_PATH", PROJECT_PATH)

    if "--settings" in sys_argv:
        index = sys_argv.index("--settings")
        SETTINGS_MODULE = sys_argv[index + 1]
    else:
        SETTINGS_MODULE = "configuration"
        # Load settings from all sources and merge them
        settings_dict = load_settings(SETTINGS_MODULE)

    # Override PORTS setting if running tests
    if "test" in sys_argv:
        settings_dict["PORTS"] = [
            False,
        ]

    # Override INSTALLED_APPS
    if (
        "BASE_INSTALLED_APPS" in settings_dict
        and "INSTALLED_APPS" in settings_dict
    ):
        settings_dict["INSTALLED_APPS"] = settings_dict[
            "BASE_INSTALLED_APPS"
        ] + list(settings_dict["INSTALLED_APPS"])
    if "REMOVED_APPS" in settings_dict:
        for app in settings_dict["REMOVED_APPS"]:
            settings_dict["INSTALLED_APPS"].remove(app)

    # Merge MIDDLEWARE
    if "BASE_MIDDLEWARE" in settings_dict and "MIDDLEWARE" in settings_dict:
        settings_dict["MIDDLEWARE"] = settings_dict["BASE_MIDDLEWARE"] + list(
            settings_dict["MIDDLEWARE"]
        )

    # Configure Django settings
    from django.conf import settings

    settings.configure(**settings_dict)

    # Set timezone
    os.environ["TZ"] = settings.TIME_ZONE

    # Handle version command
    if len(sys_argv) > 1 and sys_argv[1] in ["version", "--version"]:
        from base import get_version

        sys.stdout.write(get_version() + "\n")
        return

    # Execute Django management command
    from django.core.management import execute_from_command_line

    execute_from_command_line(sys_argv)


def entry():
    """
    Entry point for the Fiduswriter application when installed as a package.
    Uses the current working directory as the project path.
    """
    os.environ.setdefault("NO_COMPILEMESSAGES", "true")
    inner(os.getcwd())


if __name__ == "__main__":
    inner(SRC_PATH)



# Generated by Django 3.1 on 2020-08-20 03:10

from django.db import migrations, models
import usermedia.models


class Migration(migrations.Migration):
    dependencies = [
        ("usermedia", "0005_move_json_data"),
    ]

    operations = [
        migrations.AlterField(
            model_name="documentimage",
            name="copyright",
            field=models.JSONField(default=usermedia.models.default_copyright),
        ),
        migrations.AlterField(
            model_name="userimage",
            name="copyright",
            field=models.JSONField(default=usermedia.models.default_copyright),
        ),
    ]



from django.urls import re_path

from . import consumers

urlpatterns = [
    re_path(
        "^(?P<document_id>[0-9]+)/$",
        consumers.WebsocketConsumer.as_asgi(),
        name="document_consumer",
    ),
]



# Generated by Django 5.1.2 on 2024-12-10 12:00

from django.db import migrations


def article_to_doc_selector(apps, schema_editor):
    DocumentStyle = apps.get_model("style", "DocumentStyle")
    for style in DocumentStyle.objects.all():
        style.contents = style.contents.replace(".article-", ".doc-")
        style.save()


def doc_to_article_selector(apps, schema_editor):
    DocumentStyle = apps.get_model("style", "DocumentStyle")
    for style in DocumentStyle.objects.all():
        style.contents = style.contents.replace(".doc-", ".article-")
        style.save()


class Migration(migrations.Migration):
    dependencies = [
        ("style", "0001_squashed_20200219"),
    ]

    operations = [
        migrations.RunPython(
            article_to_doc_selector,
            doc_to_article_selector,
        ),
    ]



# Generated by Django 3.2.4 on 2021-07-05 13:56

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bibliography", "0002_move_json_data"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="entry",
            options={"verbose_name_plural": "Entries"},
        ),
    ]
